import os

import sqlite3
import numpy as np
import pandas as pd
import arrow as ar
from tqdm.auto import tqdm
import matplotlib.pyplot as plt


dict_growth_stages = {
    1: "Jan 1",
    2: "Jan 11",
    3: "Jan 21",
    4: "Jan 31",
    5: "Feb 10",
    6: "Feb 20",
    7: "Mar 1",
    8: "Mar 11",
    9: "Mar 21",
    10: "Mar 31",
    11: "Apr 10",
    12: "Apr 20",
    13: "Apr 30",
    14: "May 10",
    15: "May 20",
    16: "May 30",
    17: "Jun 9",
    18: "Jun 19",
    19: "Jun 29",
    20: "Jul 9",
    21: "Jul 19",
    22: "Jul 29",
    23: "Aug 8",
    24: "Aug 18",
    25: "Aug 28",
    26: "Sep 7",
    27: "Sep 17",
    28: "Sep 27",
    29: "Oct 7",
    30: "Oct 17",
    31: "Oct 27",
    32: "Nov 6",
    33: "Nov 16",
    34: "Nov 26",
    35: "Dec 6",
    36: "Dec 16",
    37: "Dec 26",
}

dict_growth_stages_biweekly = {
    1: "Jan 1",
    2: "Jan 15",
    3: "Jan 29",
    4: "Feb 12",
    5: "Feb 26",
    6: "Mar 11",
    7: "Mar 25",
    8: "Apr 8",
    9: "Apr 22",
    10: "May 6",
    11: "May 20",
    12: "Jun 3",
    13: "Jun 17",
    14: "Jul 1",
    15: "Jul 15",
    16: "Jul 29",
    17: "Aug 12",
    18: "Aug 26",
    19: "Sep 9",
    20: "Sep 23",
    21: "Oct 7",
    22: "Oct 21",
    23: "Nov 4",
    24: "Nov 18",
    25: "Dec 2",
    26: "Dec 16",
    27: "Dec 31",
}


dict_growth_stages_monthly = {
    1: "Jan 1",
    2: "Feb 1",
    3: "Mar 1",
    4: "Apr 1",
    5: "May 1",
    6: "Jun 1",
    7: "Jul 1",
    8: "Aug 1",
    9: "Sep 1",
    10: "Oct 1",
    11: "Nov 1",
    12: "Dec 1",
}


def remove_last_part(s):
    # Function to remove the part after the last underscore, including the last underscore
    # e.g. 'MIN_ESI4WK_33' will return 'MIN_ESI4WK'
    # 'TNx_33' will return 'TNx'
    return "_".join(s.split("_")[:-1])


def matplotlib_setup():
    import matplotlib.pyplot as plt

    plt.rcParams["font.family"] = "sans-serif"
    plt.rcParams["font.sans-serif"] = "Helvetica"

    # Set styles for axes
    plt.rcParams["axes.edgecolor"] = "#333F4B"
    plt.rcParams["axes.linewidth"] = 0.8
    plt.rcParams["xtick.color"] = "#333F4B"
    plt.rcParams["ytick.color"] = "#333F4B"


def delete_empty_dirs(_dir):
    """
    Cleanup by deleting folders which have no files in them. Delete folders which only have empty subdirs
    Args:
        _dir:

    Returns:

    """
    _dirs = [x[0] for x in os.walk(_dir)]

    for _d in _dirs:
        if not len([entry for entry in os.scandir(_d) if entry.is_file()]):
            try:
                os.removedirs(_d)
            except OSError:
                pass


def compute_zscore(val1, vals):
    """

    :param vals:
    :return:
    """
    import bottleneck as bn

    # Compute z-score for the values in the column `col`
    zscore = (val1 - bn.nanmean(vals)) / bn.nanstd(vals)

    return zscore


def add_sos(df, method, group_by):
    """

    :param df:
    :param method:
    :param group_by:
    :return:
    """
    groups = df.groupby(group_by, dropna=False)

    for key, vals in tqdm(groups, desc="Adding start of season info"):
        if not vals.empty:
            df.loc[vals.index, f"sos {method}"] = vals["Stage"].unique()[0]

    return df


def compute_zscores(
    df, method, group_by, value_column, num_years=-1, year_column="Harvest Year"
):
    """

    :param df:
    :param method:
    :param group_by:
    :param value_column:
    :param num_years: Number of years to consider for computing z-scores, -1 implies all years
    :return:
    """
    from heapq import nsmallest

    groups = df.groupby(group_by, dropna=False)

    for key, vals in tqdm(
        groups, desc=f"Computing z-scores {value_column} {method} {num_years} years"
    ):
        suffix = "" if num_years == -1 else f" {num_years} years"
        closest_years = (
            len(vals[year_column].unique()) if num_years == -1 else num_years
        )
        harvest_years = vals[year_column].unique()

        for year in [
            ar.now().year - 3,
            ar.now().year,
        ]:  # HACK only compute z-score for the last 3 years
            current_year = ar.now().year
            other_years = harvest_years[harvest_years != current_year]

            # Get the closest `num_years` years
            closest = nsmallest(closest_years, other_years, key=lambda x: abs(x - year))
            vals_subset = vals[vals[year_column].isin(closest)]
            vals_other = vals[vals[year_column] == year]

            # Compute z-score
            if not vals_subset.empty and not vals_other.empty:
                zscore = compute_zscore(
                    vals_other[value_column].values, vals_subset[value_column].values
                )
                df.loc[vals_other.index, f"Z-Score {value_column}{suffix}"] = zscore[0]

    return df


def categorize_zscores(df, bins, labels, cut_column, output_column):
    """

    :param df:
    :param bins:
    :param labels:
    :param cut_column:
    :param output_column:
    :return:
    """
    df.loc[:, output_column] = pd.cut(df[cut_column], bins=bins, labels=labels)

    return df


def detrend_column(df, column, group_by, detrended_column):
    """

    :param df:
    :param column:
    :param group_by:
    :param detrended_column:
    :return:
    """
    from scipy import signal

    groups = df.groupby(group_by, dropna=False)

    for key, vals in groups:
        # Drop rows where Yield is NaN
        vals = vals.dropna(subset=[column])

        # If removing values results in an empty dataframe, skip
        if not vals.empty:
            # Detrend Yield column and add to original dataframe
            detrended = signal.detrend(vals[column].values)
            df.loc[vals.index, detrended_column] = detrended

    return df


def categorize_column(
    df, column, group_by, zscore_column, categories, bins, category_column
):
    """
    HACk: Needs to be updated
    :param df:
    :param column:
    :param group_by:
    :param zscore_column:
    :param categories:
    :param bins:
    :param category_column:
    :return:
    """
    groups = df.groupby(group_by, dropna=False)

    for key, vals in groups:
        # Compute z-score for the values in the Detrended Yield column
        # and add to original dataframe
        zscore = compute_zscore(vals[column].values)
        df.loc[vals.index, zscore_column] = zscore

    # Categorize the z-scores
    categorize_zscores(
        df,
        bins,
        categories,
        cut_column=zscore_column,
        output_column=category_column,
    )

    return df


def get_crop_season(filename):
    """
    Get crop name and season from filename
    """
    if "winter_wheat" in filename:
        crop = "winter_wheat"
        season_index = 1
    elif "spring_wheat" in filename:
        crop = "spring_wheat"
        season_index = 1
    elif "maize" in filename:
        crop = "maize"
    elif "rice" in filename:
        crop = "rice"
    elif "soybean" in filename:
        crop = "soybean"
    elif "sorghum" in filename:
        crop = "sorghum"
    elif "millet" in filename:
        crop = "sorghum"
    elif "teff" in filename:
        crop = "teff"
    else:
        raise ValueError(f"Crop not found in {filename}")

    if "s1" in filename:
        season_index = 1
    elif "s2" in filename:
        season_index = 2
    else:
        season_index = 1

    return crop, season_index


def create_output_directory(method, admin_zone, country, crop, path_output):
    """

    :param method:
    :param admin_zone:
    :param country:
    :param crop:

    :return:
    """
    dir_output = path_output / "cei" / "indices" / method / admin_zone / country / crop
    os.makedirs(dir_output, exist_ok=True)

    return dir_output


def to_db(db_path, table_name, df):
    """

    Args:
        db_path:
        table_name:
        df:

    Returns:

    """
    from pangres import upsert
    from sqlalchemy import create_engine

    try:
        engine = create_engine("sqlite:///" + str(db_path))

        upsert(
            con=engine,
            df=df,
            table_name=table_name,
            if_row_exists="update",
            chunksize=20,
        )
    except Exception as e:
        print(f"Exception: {e}")


def is_table(database, table_name):
    """
    Check if table_name exists in database. Return True if it does and False if not
    Args:
        database:
        table_name:

    Returns:

    """
    con = sqlite3.connect(database)

    query = "SELECT * FROM sqlite_master"
    df = pd.read_sql_query(query, con)
    con.close()

    if table_name in df["tbl_name"].values:
        return True
    else:
        return False


def plot(X, labels, probabilities=None, parameters=None, ground_truth=False, ax=None):
    if ax is None:
        _, ax = plt.subplots(figsize=(10, 4))
    labels = labels if labels is not None else np.ones(X.shape[0])
    probabilities = probabilities if probabilities is not None else np.ones(X.shape[0])
    # Black removed and is used for noise instead.
    unique_labels = set(labels)
    colors = [plt.cm.Spectral(each) for each in np.linspace(0, 1, len(unique_labels))]
    # The probability of a point belonging to its labeled cluster determines
    # the size of its marker
    proba_map = {idx: probabilities[idx] for idx in range(len(labels))}
    for k, col in zip(unique_labels, colors):
        if k == -1:
            # Black used for noise.
            col = [0, 0, 0, 1]

        class_index = np.where(labels == k)[0]
        for ci in class_index:
            try:
                ax.plot(
                    X[ci, 0],
                    X[ci, 1],
                    "x" if k == -1 else "o",
                    markerfacecolor=tuple(col),
                    markeredgecolor="k",
                    markersize=4 if k == -1 else 1 + 5 * proba_map[ci],
                )
            except:
                breakpoint()
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    preamble = "True" if ground_truth else "Estimated"
    title = f"{preamble} number of clusters: {n_clusters_}"
    if parameters is not None:
        parameters_str = ", ".join(f"{k}={v}" for k, v in parameters.items())
        title += f" | {parameters_str}"
    ax.set_title(title)
    plt.tight_layout()
    plt.show()


def pairwise_rmse(df1, df2):
    return np.sqrt(((df1 - df2) ** 2).mean())


def nse(observed, simulated, weights=None):
    """
    Compute Weighted Nash-Sutcliffe Efficiency
    :param observed: Array of observed values
    :param simulated: Array of simulated values
    :param weights: Optional array of weights
    :return: NSE value
    """
    if weights is None:
        weights = np.ones_like(observed)

    weighted_sq_diff = np.sum(weights * (observed - simulated) ** 2)
    weighted_variance = np.sum(
        weights * (observed - np.average(observed, weights=weights)) ** 2
    )

    return 1 - (weighted_sq_diff / weighted_variance)


def mape(observed, simulated, weights=None):
    """
    Compute Weighted Mean Absolute Percentage Error
    :param observed: Array of observed values
    :param forecast: Array of forecast values
    :param weights: Optional array of weights
    :return: MAPE value
    """
    if weights is None:
        weights = np.ones_like(observed)

    weighted_abs_percent_error = weights * np.abs((observed - simulated) / observed)

    return np.sum(weighted_abs_percent_error) / np.sum(weights) * 100


def pbias(observed, simulated, weights=None):
    """
    Compute Weighted Percent Bias
    :param observed: Array of observed values
    :param simulated: Array of simulated values
    :param weights: Optional array of weights
    :return: PBIAS value
    """
    if weights is None:
        weights = np.ones_like(observed)

    weighted_diff_sum = np.sum(weights * (observed - simulated))
    weighted_obs_sum = np.sum(weights * observed)

    return (weighted_diff_sum / weighted_obs_sum) * 100


# Function to remove trend
def detrend(data):
    return data.diff().dropna()


# Function to add trend back
def retrend(original_data, detrended_data):
    try:
        retrended_data = detrended_data.cumsum() + original_data.iloc[0]
    except:
        breakpoint()

    return retrended_data


def linregress(x, y):
    """ """
    # Fit a linear regression model using statsmodels
    import statsmodels.api as sm

    x = sm.add_constant(x)
    model = sm.OLS(y, x).fit()

    # Get the slope, intercept, p-value of the trendline
    intercept = model.params[0]
    slope = model.params[1]
    p = model.f_pvalue

    if p < 0.05:
        # Significant, therefore add * to the equation
        eqn = f"y = {slope:.3f}x + {intercept:.2f} *"
    else:
        eqn = f"y = {slope:.3f}x + {intercept:.2f}"

    return eqn


def slope(x, y):
    """ """
    from scipy.stats import mstats
    import pymannkendall as mk

    # Reset index to ensure x and y are aligned
    x = x.reset_index(drop=True)
    y = y.reset_index(drop=True)

    # Note that we are getting slope from one library and intercept from another
    # This is because pymannkendall reports intercept of the Kendall-Theil Robust Line
    # and theilslopes reports intercept of the Theil-Sen estimator, we want the latter
    try:
        trend, h, p, z, Tau, s, var_s, slope, intercept = mk.original_test(y)
        slope, intercept = mstats.theilslopes(y, x)[0], mstats.theilslopes(y, x)[1]
    except:
        slope = np.nan
        intercept = np.nan
        p = np.nan

    return slope, intercept, p


def is_trending(x, y, threshold=0.05):
    """ """
    import pymannkendall as mk

    # Reset index to ensure x and y are aligned
    x = x.reset_index(drop=True)
    y = y.reset_index(drop=True)

    trend, h, p, z, Tau, s, var_s, slope, intercept = mk.original_test(y)

    if p < threshold:
        return True
    else:
        return False


def process_subsets(df, custom_function, **kwargs):
    """
    Processes subsets of the dataframe based on unique values in two columns and
    applies a given custom function

    :param df (pd.DataFrame): The dataframe to process
    :param custom_function: A function to apply to each subsubset dataframe
    :param **kwargs  Additional keyword arguments to pass to the operation function
    """
    frames = []
    for i, value1 in enumerate(df[kwargs["column1"]].unique()):
        subset_df = df[df[kwargs["column1"]] == value1]

        if not subset_df.empty:
            if kwargs["column2"]:
                for j, value2 in enumerate(subset_df[kwargs["column2"]].unique()):
                    subsubset_df = subset_df[subset_df[kwargs["column2"]] == value2]

                    results = custom_function(subsubset_df, i, j, **kwargs)
                    frames.append(results)
            else:
                results = custom_function(subsubset_df, i, i, **kwargs)
                frames.append(results)

    df_results = pd.concat(frames)

    return df_results


def compute_dataframe_transformation(df, group_by=[], column=None, stat="mean"):
    """

    :param df:
    :param group_by:
    :param column:
    :param stat:
    """
    if stat == "mean":
        df = df.groupby(group_by, dropna=False)[column].mean().reset_index()
    else:
        raise NotImplementedError(f"stat {stat} not implemented")

    # Drop any rows where column is NaN
    df = df.dropna(subset=[column])

    return df


def list_directories(directory_path):
    """
    Lists all directories within the specified directory path.

    Parameters:
    - directory_path: A string representing the path to the directory you want to explore.

    Returns:
    - A list of directory names found within the specified directory.
    """
    try:
        # List all entries in the directory given by "directory_path"
        directory_contents = os.listdir(directory_path)

        # Filter out the directories from all entries
        directories = [
            d
            for d in directory_contents
            if os.path.isdir(os.path.join(directory_path, d))
        ]

        return directories
    except FileNotFoundError:
        print(f"Error: The directory '{directory_path}' was not found.")
        return []
    except PermissionError:
        print(f"Error: Permission denied to access the directory '{directory_path}'.")
        return []
    except Exception as e:
        print(f"An error occurred: {e}")
        return []


def compute_biweekly_index(row):
    """
    Compute the index of the biweekly period for a given day of the year.

    The first biweekly period of the year starts on January 1st.
    Args:
    - row (pd.Series): A row from the DataFrame containing 'year' and 'doy' columns.

    Returns:
    - int: The biweekly period index (starting from 1).
    """
    # Calculate the day of the year, adjusting to start from 0
    day_of_year_zero_indexed = int(row["Doy"]) - 1
    # Compute the biweekly index, adjusting so the first period is index 1
    biweekly_index = (day_of_year_zero_indexed // 14) + 1

    return biweekly_index


def is_leap_year(year):
    """Check if the specified year is a leap year."""
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)


def compute_time_periods(df_inp, period_type, year):
    """
    Compute the starting and ending time-periods for a given period type ('dekad', 'biweekly', 'monthly')
    in either a leap or non-leap year.

    Parameters:
    - df: DataFrame with a 'Doy' column.
    - period_type: String indicating the period type ('dekad', 'biweekly', 'monthly').
    - year: The year for the Doy values, used to handle leap years.

    Returns:
    - A tuple with starting and ending periods for the specified period type.
    """
    # Adjust the origin based on whether the year is a leap year
    df = df_inp.copy()
    df["Date"] = pd.to_datetime(
        df["Doy"], unit="D", origin=pd.Timestamp(f"{year}-01-01")
    )

    if period_type.startswith("dekad"):
        df["Period"] = ((df["Date"].dt.dayofyear - 1) // 10) + 1
    elif period_type.startswith("biweekly"):
        df["Period"] = ((df["Date"].dt.dayofyear - 1) // 14) + 1
    elif period_type.startswith("monthly"):
        df["Period"] = df["Month"]
    else:
        return "Invalid period type. Choose 'dekad', 'biweekly', or 'monthly'."

    start_period = df.iloc[0]["Period"]
    end_period = df.iloc[-1]["Period"]

    return start_period, end_period


def compute_h_index(values):
    # Sort the array in descending order
    sorted_value = np.sort(values)[::-1]

    # Iterate through the sorted array to find the h-index
    h_index = 0

    for i, value in enumerate(sorted_value, start=1):
        if value >= i:
            h_index = value
        else:
            break

    return h_index


def get_z_value(alpha):
    """
    Calculate the z-value for a given alpha level.

    Parameters:
    alpha (float): The significance level (e.g., 0.05 for a 95% confidence interval)

    Returns:
    float: The corresponding z-value
    """
    from scipy.stats import norm

    return norm.ppf(1 - alpha / 2)



import geopandas as gpd
import pandas as pd
import matplotlib.pyplot as plt
import palettable as pal
import matplotlib.colors as mcolors

import pandas as pd
import glob
import os

# 1. Specify the directory containing your .dta files:
data_dir = r"C:\Users\ritvik\Downloads\maize_yield\maize_yield"

# 2. Use glob to find all .dta files in that directory:
dta_files = glob.glob(os.path.join(data_dir, "*.dta"))

# 3. Read each .dta file into a pandas DataFrame and store in a list:
dataframes = [pd.read_stata(f) for f in dta_files]

# 4. Concatenate them all into one DataFrame (row-wise):
merged_df = pd.concat(dataframes, ignore_index=True)

merged_df['ZONE'] = merged_df['ZONE'].astype(int)
merged_df['DIST'] = merged_df['DIST'].astype(int)

# create a column called W_CODE which is set up as follows
# create a string by converting ZONE column to string and append 0
# to the left of the string to make it 2 characters long
# then do the same with DIST column
# finally concatenate the two strings
merged_df['W_CODE'] = merged_df['ZONE'].astype(str).str.zfill(2) + merged_df['DIST'].astype(str).str.zfill(2)

merged_df['W_CODE'] = '7' + merged_df['W_CODE']

# Remove the .0 at the end of the string in W_CODE
merged_df['W_CODE'] = merged_df['W_CODE'].str.replace('.0', '')
merged_df['W_CODE'] = merged_df['W_CODE'].astype(int)

dg = gpd.read_file(r"D:\Users\ritvik\projects\GEOGLAM\Input\countries\wolayita\wolayita_dissolved.shp")
dg = dg[['W_CODE', 'W_NAME']]

# Merge the two dataframes on W_CODE
merged_df = pd.merge(merged_df, dg, on='W_CODE', how='left')

# Remove rows where PROD98CQ or AREAH are null
merged_df = merged_df.dropna(subset=['PROD98CQ', 'AREAH'])

# Compte yield column
merged_df['yield'] = merged_df['PROD98CQ'] / merged_df['AREAH']

# create a new dataframe which computes average yield by W_NAME for each year
df_avg_yield = merged_df.groupby(['W_NAME', 'YEAR'])['yield'].mean().reset_index()

# Change W_NAME column to title case
df_avg_yield['W_NAME'] = df_avg_yield['W_NAME'].str.title()

# Change YEAR to int
df_avg_yield['YEAR'] = df_avg_yield['YEAR'].astype(int)

# Convert to a format where each YEAR is converted to int and becomes a column and yield is the value
df_avg_yield = df_avg_yield.pivot(index='W_NAME', columns='YEAR', values='yield')

# Remove YEAR as column name and W_NAME as index name
df_avg_yield.index.name = None
df_avg_yield.columns.name = None

df_avg_yield.to_csv('wolayita_yields.csv')

breakpoint()
# 5. (Optional) Inspect the merged DataFrame
print(merged_df.head())
print(len(merged_df))
merged_df.to_csv('merged_df.csv', index=False)
breakpoint()

import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

import geopandas as gpd
dg = gpd.read_file(r"D:\Users\ritvik\projects\GEOGLAM\wolayita.shp")
dg = dg[dg['Z_NAME'] == "Wolayita"]

# Dissolve on W_NAME column
dg = dg.dissolve(by="W_NAME")

# save to disk
dg.to_file(r"D:\Users\ritvik\projects\GEOGLAM\Input\countries\wolayita\wolayita_dissolved.shp")

breakpoint()
# 1. Load the dataset
df = pd.read_csv('merged_df.csv')

# 2. Ensure we have a 'yield' column.
#    If not present, we compute yield as Maize_Production / Maize_Area.
if 'yield' not in df.columns:
    if 'PROD98CQ' in df.columns and 'AREAH' in df.columns:
        # Compute yield in tonnes per hectare (or adjust unit if needed)
        df['yield'] = df['PROD98CQ'] / df['AREAH']
    else:
        raise ValueError("The required columns to compute yield are missing.")

# 3. Calculate percentage of missing data for yield
missing_pct_yield = df['yield'].isnull().mean() * 100
print(f"Percentage of missing data for yield: {missing_pct_yield:.2f}%")

# 4. Check if some years have more or less data
#    Count the number of records for each year
year_counts = df['YEAR'].value_counts().sort_index()
print("\nNumber of records per year:")
print(year_counts)

# 5. Plot histogram of yield distributions by year
import seaborn as sns

# Instead of looping and plotting histograms, we can use a boxplot
plt.figure(figsize=(12, 8))

sns.boxplot(x='YEAR', y='yield', data=df)

# Add labels and title
plt.xlabel("")
plt.ylabel("Yield")

plt.show()


# Group by YEAR and get size (number of rows)
df_year_counts = df.groupby('YEAR').size().reset_index(name='Count')
# Sort by YEAR if you want ascending year order
df_year_counts.sort_values(by='YEAR', inplace=True)

plt.figure(figsize=(10, 6))
sns.barplot(data=df_year_counts, x='YEAR', y='Count', color='skyblue', edgecolor='black')

plt.xlabel("")
plt.ylabel("Number of Yield Records")
plt.xticks(rotation=45)  # Rotate x labels if needed
plt.tight_layout()       # Adjust layout to avoid clipping
plt.show()


import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

# 1. Group by FA and YEAR, then calculate the mean yield
fa_year_yield = df.groupby(['FA', 'YEAR'])['yield'].mean().reset_index()

# 2. Pivot so rows = FA, columns = YEAR, values = average yield
fa_year_pivot = fa_year_yield.pivot(index='FA', columns='YEAR', values='yield')

# 3. Create the heatmap
plt.figure(figsize=(12, 8))
sns.heatmap(
    fa_year_pivot,
    cmap='viridis',       # color map; try 'coolwarm' or others
    annot=False,           # show numeric values in each cell
    fmt=".2f",            # format numbers (2 decimal places)
    linewidths=.5         # line width between cells
)

plt.title("Heatmap of Average Yield by FA and YEAR")
plt.xlabel("YEAR")
plt.ylabel("FA")
plt.tight_layout()
plt.show()

breakpoint()


# --- Read and preprocess your main shapefile ---
dg = gpd.read_file(r"D:\Users\ritvik\projects\GEOGLAM\safrica.shp")

# remove rows where both ADMIN1 and ADMIN2 are null
dg = dg.dropna(subset=["ADMIN1", "ADMIN2"], how="all")

# if ADMIN2 is not null then replace ADMIN1 with ADMIN2 values
dg["ADMIN1"] = dg["ADMIN2"].combine_first(dg["ADMIN1"])

# --- Read your CSV and merge on ADMIN1 ---
df = pd.read_csv(r"C:\Users\ritvik\Downloads\geocif.csv")

dg = dg.merge(
    df[["ADMIN1", 'Predicted Yield (tn per ha)',
        'Median Yield (tn per ha) (2013-2017)', 'Predicted/Median']],
    on="ADMIN1",
    how="left"
)

# --- Create a dissolved national boundary GeoDataFrame ---
boundary_gdf = dg.dissolve(by="ADMIN0")

# --- Colormap and normalization setup ---
cmap = pal.colorbrewer.get_map("BrBG", "diverging", 11).mpl_colormap
norm = mcolors.TwoSlopeNorm(vmin=-40, vcenter=0, vmax=40)

# --- First map: Predicted/Median ---
fig, ax = plt.subplots(figsize=(10, 6))

# 1) Plot the main layer
dg.plot(
    column="Predicted/Median",
    cmap=cmap,
    norm=norm,
    legend=True,
    ax=ax,
    edgecolor='gray',
    linewidth=0.2,
    legend_kwds={
        "shrink": 0.5,
        "pad": 0.002,
        "orientation": "horizontal"
    }
)

url = "https://naciscdn.org/naturalearth/110m/cultural/ne_110m_admin_0_countries.zip"

world = gpd.read_file(url)
world =  world[world['ADMIN'].isin(['South Africa', 'Angola', 'Malawi', 'Zambia'])]

# 2) Plot the dissolved national boundaries on top
world.plot(
    ax=ax,
    color="none",       # No fill
    edgecolor="black",  # Outline color
    linewidth=0.5
)

ax.set_title("Maize Yield Forecast % Anomaly")
plt.axis("off")
plt.tight_layout()
plt.savefig("aa.png", dpi=300)
plt.close()


# --- Second map: Median Yield (2013-2017) ---
# fig, ax = plt.subplots(figsize=(10, 6))
#
# # 1) Plot the main layer
# dg.plot(
#     column="Median Yield (tn per ha) (2013-2017)",
#     cmap=cmap,
#     legend=True,
#     ax=ax,
#     legend_kwds={
#         "shrink": 0.5,
#         "pad": 0.002,
#         "orientation": "horizontal"
#     }
# )
#
# # 2) Plot the dissolved national boundaries on top
# boundary_gdf.plot(
#     ax=ax,
#     color="none",
#     edgecolor="black",
#     linewidth=1
# )
#
# ax.set_title("Median Maize Yield (2013-2017)")
# plt.axis("off")
# plt.tight_layout()
# plt.show()
# plt.close()

breakpoint()



# panel serve climate_risk_dashboard.py
import ast
import os
from configparser import ConfigParser
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

import wandb
import arrow as ar
import geopandas as gp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm import tqdm

from geocif import logger as log
from geocif import utils
from .cei import definitions as di
from .ml import correlations
from .ml import spatial_autocorrelation as sa
from .ml import feature_engineering as fe
from .ml import feature_selection as fs
from .ml import output
from .ml import stages
from .ml import stats
from .ml import trainers
from .ml import trend
from .ml import xai

plt.style.use("default")

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)


@dataclass
class Geocif:
    method: str = "dekad_r"
    group_by: List[str] = field(
        default_factory=lambda: ["Index", "Country", "Region", "Crop", "Season"]
    )
    metrics: List[str] = field(default_factory=lambda: ["$r^2$", "RMSE", "MAE", "MAPE"])
    logger: log = None
    parser: ConfigParser = field(default_factory=ConfigParser)
    project_name: str = "geocif"

    def __post_init__(self):
        self.dir_output = Path(self.parser.get("PATHS", "dir_output"))
        self.dir_condition = Path(self.parser.get("PATHS", "dir_condition"))
        os.makedirs(self.dir_output / "ml", exist_ok=True)

        self.country: str = None
        self.crop: str = None
        self.forecast_season: int = None
        self.all_stages: list = []
        self.all_seasons: list = []
        self.all_seasons_with_yield: list = []
        self.model_names: list = []
        self.feature_names: list = []
        self.selected_features: list = []

        self._date = ar.utcnow().to("America/New_York")
        self.today = self._date.format("MMMM_DD_YYYY")
        self.today_year = self._date.year
        self.today_doy = int(self._date.format("DDD"))
        self.today_full = self._date.format("MMMM_DD_YYYY_HH_mm")

        self.df_forecast = pd.DataFrame()
        """
        ====================================================================
                                        Config file: Logging
        ====================================================================
        """
        self.log_level = self.parser.get("LOGGING", "log_level")

        """
        ====================================================================
                                Config file: Default
        ====================================================================
        """
        self.method = self.parser.get("DEFAULT", "method")
        self.db_forecasts = self.parser.get("DEFAULT", "db")
        self.countries = ast.literal_eval(self.parser.get("DEFAULT", "countries"))
        self.do_parallel = self.parser.getboolean("DEFAULT", "do_parallel")
        self.update_input_file = self.parser.getboolean("DEFAULT", "update_input_file")
        self.correlation_plots = self.parser.getboolean("DEFAULT", "correlation_plots")
        self.national_correlation = self.parser.getboolean(
            "DEFAULT", "national_correlation"
        )
        self.plot_map_for_correlation_plot = self.parser.getboolean(
            "DEFAULT", "plot_map_for_correlation_plot"
        )
        self.correlation_plot_groupby = self.parser.get(
            "DEFAULT", "correlation_plot_groupby"
        )
        self.run_ml = self.parser.getboolean("DEFAULT", "run_ml")

        """
        ====================================================================
                                Config file: ML
        ====================================================================
        """
        self.model_type = self.parser.get("ML", "model_type")
        self.classify_target = self.parser.getboolean("ML", "classify_target")
        self.number_classes = self.parser.getint("ML", "number_classes")
        self.target = self.parser.get("ML", "target")
        self.rename_target = self.parser.getboolean("ML", "rename_target")
        self.new_name_target = self.parser.get("ML", "new_name_target")
        self.fraction_simulate = self.parser.getint("ML", "fraction_simulate")
        self.analogous_year_yield_as_feature = self.parser.getboolean(
            "ML", "analogous_year_yield_as_feature"
        )
        self.correlation_threshold = self.parser.getfloat("ML", "correlation_threshold")
        self.include_lat_lon_as_feature = self.parser.getboolean(
            "ML", "include_lat_lon_as_feature"
        )
        self.spatial_autocorrelation = self.parser.getboolean(
            "ML", "spatial_autocorrelation"
        )
        self.sa_method = self.parser.get("ML", "sa_method")
        self.last_year_yield_as_feature = self.parser.getboolean(
            "ML", "last_year_yield_as_feature"
        )
        self.panel_model = self.parser.getboolean("ML", "panel_model")
        self.panel_model_region = self.parser.get("ML", "panel_model_region")
        self.use_outlook_as_feature = self.parser.getboolean(
            "ML", "use_outlook_as_feature"
        )
        self.use_single_time_period_as_feature = self.parser.getboolean(
            "ML", "use_single_time_period_as_feature"
        )
        self.lag_yield_as_feature = self.parser.getboolean("ML", "lag_yield_as_feature")
        self.number_median_years = self.parser.getint("ML", "median_years")
        self.median_yield_as_feature = self.parser.getboolean(
            "ML", "median_yield_as_feature"
        )
        self.median_area_as_feature = self.parser.getboolean(
            "ML", "median_area_as_feature"
        )
        self.number_lag_years = self.parser.getint("ML", "lag_years")
        self.cluster_strategy = self.parser.get("ML", "cluster_strategy")
        self.feature_selection = self.parser.get("ML", "feature_selection")
        self.check_yield_trend = self.parser.getboolean("ML", "check_yield_trend")
        self.run_latest_time_period = self.parser.getboolean(
            "ML", "run_latest_time_period"
        )
        self.run_every_time_period = self.parser.get("ML", "run_every_time_period")
        self.cat_features: list = ast.literal_eval(
            self.parser.get("ML", "cat_features")
        )

        self.use_cumulative_features = self.parser.getboolean(
            "DEFAULT", "use_cumulative_features"
        )
        """
        ====================================================================
                                Variables, Paths
        ====================================================================
        """
        # If ML model is run for individual region or cluster, then Region_ID is the same for each region
        # or cluster and therefore redundant for the ML model
        # if self.cluster_strategy in ["individual", "auto_detect"]:
        #    self.cat_features.remove("Region_ID")

        self.target_bins = {}

        self.fixed_columns: list = [
            "Country",
            "Region",
            "Crop",
            "Area",
            "Season",
            "Harvest Year",
        ]
        self.target: str = "Yield (tn per ha)"
        self.statistics_columns: list = [
            "Area (ha)",
            "Production (tn)",
        ]

        self.combined_dict = {
            **di.dict_indices,
            **di.dict_ndvi,
            **di.dict_gcvi,
            **di.dict_esi4wk,
            **di.dict_hindex,
        }

        self.combined_keys = list(self.combined_dict.keys())

        self.dir_ml = self.dir_output / "ml"
        self.dir_db = self.dir_ml / "db"
        self.dir_analysis = self.dir_ml / "analysis" / self.today
        dir_input = Path(self.parser.get("PATHS", "dir_input"))
        self.dir_shapefiles = dir_input / "Global_Datasets" / "Regions" / "Shps"

        os.makedirs(self.dir_db, exist_ok=True)
        os.makedirs(self.dir_analysis, exist_ok=True)

        self.db_path = self.dir_db / self.db_forecasts

        # self.pickle_file = self.base_dir / self.parser.get("outlook", "pickle_file")
        # obj_pickle = outlook.Outlook(self.pickle_file)
        # self.df_outlook = obj_pickle.read_outlook_file()

    def train(self, df_region, scaler=None):
        """

        Args:
            df_region:
            scaler:

        Returns:

        """

        """ Perform feature selection """
        if self.model_type == "REGRESSION":
            target_column = (
                f"Detrended {self.target}" if self.check_yield_trend else self.target
            )
        elif self.model_type == "CLASSIFICATION":
            target_column = self.target_class

        # Drop rows where target_column is NaN
        df_region = df_region.dropna(subset=[target_column])

        X_train = df_region[self.feature_names]
        # Drop any columns with NaNs
        X_train = X_train.dropna(axis=1, how="any")
        y_train = df_region[target_column]

        if self.ml_model:
            if self.model_name in ["cumulative_1", "cumulative_2", "cumulative_3"]:
                all_features = X_train.columns

                # Select the columns with use_ceis in it
                self.selected_features = [
                    column
                    for column in all_features
                    if any(cei in column for cei in self.use_ceis)
                ]
            else:
                # self.logger.info(f"Selecting features for {self.country} {self.crop}")
                selector, _, self.selected_features = fs.select_features(
                    X_train, y_train, method=self.feature_selection
                )
                # self.logger.info(f"Selected features: {self.selected_features}")

            """ Update model to include conformal estimates """
            if "lat" not in self.selected_features and self.include_lat_lon_as_feature:
                self.selected_features.append("lat")
            if "lon" not in self.selected_features and self.include_lat_lon_as_feature:
                self.selected_features.append("lon")
            X_train = df_region[self.selected_features + self.cat_features]

            dir_output = (
                self.dir_analysis
                / self.country
                / self.crop
                / self.model_name
                / str(self.forecast_season)
            )

            region_id = df_region["Region_ID"].unique()[0]
            X_train.to_csv(dir_output / f"X_train_{region_id}.csv", index=False)
            if scaler:
                X_train_nocat = X_train.drop(
                    columns=[
                        item for item in self.cat_features if item != "Harvest Year"
                    ]
                )
                X_train_scaled = scaler.fit_transform(X_train_nocat)
            else:
                X_train_scaled = X_train

            """ Train model """
            self.best_hyperparams, self.model = trainers.auto_train(
                self.cluster_strategy,
                self.model_name,
                self.model_type,
                False,
                "Harvest Year",
                df_region[self.selected_features + self.cat_features + [self.target]],
                X_train_scaled,
                y_train,
                feature_names=self.selected_features,
                target_col=target_column,
                optimize=self.optimize,
                fraction_loocv=self.fraction_loocv,
                cat_features=self.cat_features,
            )

            """ Estimate CI only if flag is True """
            if self.estimate_ci:
                if self.estimate_ci_for_all or self.forecast_season == self.today_year:
                    self.model = trainers.estimate_ci(
                        self.model_type, self.model_name, self.model
                    )

            try:
                if self.model_name == "catboost":
                    self.model.fit(
                        X_train,
                        y_train,
                        cat_features=self.cat_features,
                        verbose=False,
                    )
                elif self.model_name in ["ngboost", "oblique", "tabpfn"]:
                    X_train = X_train.drop(
                        columns=[
                            item for item in self.cat_features if item != "Harvest Year"
                        ]
                    )
                    self.model.fit(X_train, y_train)
                elif self.model_name == "ydf":
                    # Combine X_train and y_train
                    df_train = pd.concat([X_train, y_train], axis=1)

                    self.model = self.model.train(df_train)
                elif self.model_name == "geospaNN":
                    self.model.fit(
                        X_train,
                        y_train,
                        # callbacks=[TQDMCallback(self.best_hyperparams["iterations"])],
                    )
                elif self.model_name == "merf":
                    Z_train = np.ones((len(X_train), 1))
                    clusters_train = df_region["Region"]
                    clusters_train.reset_index(drop=True, inplace=True)

                    self.model.fit(
                        X_train,
                        Z_train,
                        clusters_train.astype("object"),
                        y_train.values,
                    )
                elif self.model_name == "linear":
                    self.model.fit(X_train_scaled, y_train)
                elif self.model_name == "gam":
                    self.model.fit(X_train_scaled.values, y_train.values)
                    self.best_hyperparams = {}
                elif self.model_name in ["cubist"]:
                    self.model.fit(X_train, y_train)
                elif self.model_name in [
                    "cumulative_1",
                    "cumulative_2",
                    "cumulative_3",
                ]:
                    from sklearn.preprocessing import StandardScaler, LabelEncoder

                    if self.model_name == "cumulative_1":
                        num_columns = 1
                    elif self.model_name == "cumulative_2":
                        num_columns = 2
                    elif self.model_name == "cumulative_3":
                        num_columns = 3

                    # Standardize the numeric features
                    scaler = StandardScaler()
                    X_numeric = X_train.iloc[:, :num_columns]
                    X_scaled_numeric = pd.DataFrame(
                        scaler.fit_transform(X_numeric),
                        columns=X_numeric.columns,
                        index=X_train.index,
                    )

                    # Encode the Region as categorical
                    le = LabelEncoder()
                    X_region = pd.Series(
                        le.fit_transform(X_train["Region"]),
                        name="Region",
                        index=X_train.index,
                    )

                    # Combine scaled numeric features and encoded region
                    X_train_scaled = pd.concat([X_scaled_numeric, X_region], axis=1)

                    self.model.fit(X_train_scaled, y_train)
            except Exception as e:
                self.logger.error(
                    f"Error fitting model for {self.country} {self.crop} {e}"
                )

    def predict(self, df_region, scaler=None):
        """
        Predict yield for the current stage
        :param df_region:
        :param scaler:
        """
        """ Select dataframe for prediction"""
        X_test = df_region[self.selected_features + self.cat_features]
        y_test = df_region[self.target].values
        best_hyperparameters = {}

        """ Do prediction based on model type"""
        if not self.ml_model:
            best_hyperparameters = np.nan
            if self.model_name == "analog":
                y_pred = np.full(len(X_test), df_region["Analogous Year Yield"].values)
            elif self.model_name == "median":
                y_pred = np.full(len(X_test), df_region[f"Median {self.target}"].values)
            elif self.model_name == "last_year":
                y_pred = np.full(
                    len(X_test), df_region[f"Last Year {self.target}"].values
                )
        else:
            if self.model_name in ["linear", "gam"]:
                # Drop cat_features from X_test
                X_test = X_test.drop(
                    columns=[
                        item for item in self.cat_features if item != "Harvest Year"
                    ]
                )
                X_test = scaler.transform(X_test)
                best_hyperparameters = {}

            if self.estimate_ci:
                if self.estimate_ci_for_all or self.forecast_season == self.today_year:
                    if self.model_name in ["logistic", "catboost"] and self.model_type == "CLASSIFICATION":
                        y_pred = self.model.predict(X_test)
                        y_pred_ci = self.model.predict_proba(X_test)
                    elif self.model_name == "ngboost":
                        y_pred = self.model.predict(X_test)

                        if self.model_type == "REGRESSION":
                            y_dists = self.model.pred_dist(X_test)
                            z_value = utils.get_z_value(self.alpha)

                            # Extract means and standard deviations from the distribution predictions
                            means = y_dists.loc
                            std_devs = y_dists.scale

                            # Calculate lower and upper bounds of the prediction intervals
                            lower_bounds = means - z_value * std_devs
                            upper_bounds = means + z_value * std_devs

                            # Output would be similar to Mapie: point predictions, lower bounds, and upper bounds
                            y_pred_ci = np.vstack([lower_bounds, means, upper_bounds]).T
                        elif self.model_type == "CLASSIFICATION":
                            y_pred_ci = self.model.predict_proba(X_test)

                            # Output should be similar to Mapie: point predictions, lower bounds, and upper bounds
                            y_pred_ci = np.vstack(
                                [y_pred_ci[:, 0], y_pred, y_pred_ci[:, 1]]
                            ).T
                    else:
                        y_pred, y_pred_ci = self.model.predict(X_test, alpha=self.alpha)
                        best_hyperparameters = self.model.get_params().copy()
                        # If estimator key is in best_hyperparameters, remove it
                        if "estimator" in best_hyperparameters:
                            del best_hyperparameters["estimator"]
            elif self.model_name == "merf":
                Z_test = np.ones((len(X_test), 1))
                clusters_test = df_region["Region"]
                clusters_test.reset_index(drop=True, inplace=True)

                y_pred = self.model.predict(
                    X_test, Z_test, clusters_test.astype("object")
                )
                best_hyperparameters = self.model.fe_model.get_params().copy()
            elif self.model_name in ["cumulative_1", "cumulative_2", "cumulative_3"]:
                from sklearn.preprocessing import StandardScaler, LabelEncoder

                if self.model_name == "cumulative_1":
                    num_columns = 1
                elif self.model_name == "cumulative_2":
                    num_columns = 2
                elif self.model_name == "cumulative_3":
                    num_columns = 3

                # Standardize the numeric features
                scaler = StandardScaler()
                X_numeric = X_test.iloc[:, :num_columns]
                try:
                    X_scaled_numeric = pd.DataFrame(
                        scaler.fit_transform(X_numeric),
                        columns=X_numeric.columns,
                        index=X_test.index,
                    )
                except:
                    breakpoint()

                # Encode the Region as categorical
                le = LabelEncoder()
                X_region = pd.Series(
                    le.fit_transform(X_test["Region"]),
                    name="Region",
                    index=X_test.index,
                )

                # Combine scaled numeric features and encoded region
                X_test_scaled = pd.concat([X_scaled_numeric, X_region], axis=1)
                y_pred = self.model.predict(X_test_scaled)
                best_hyperparameters = {}  # self.model.get_params().copy()
            elif self.model_name == "geospaNN":
                import torch
                import geospaNN

                # Remove any categorical features
                X_test = X_test.drop(columns=self.cat_features)
                X = torch.from_numpy(X_test.to_numpy()).float()
                coord = torch.from_numpy(
                    self.df_test[["lon", "lat"]].to_numpy()
                ).float()

                p = X.shape[1]
                n = X.shape[0]
                nn = 5

                data = geospaNN.make_graph(X, Y, coord, nn)

                # remove categorical features from df_train
                data_train = df_region[
                    self.selected_features + self.cat_features + [self.target]
                ]
                w_train = data_train.y - self.estimate(data_train.x)
            elif self.model_name == "ydf":
                y_pred = self.model.evaluate(X_test)
                best_hyperparameters = {}
            elif self.model_name == "tabpfn":
                X_test = X_test.drop(
                    columns=[
                        item for item in self.cat_features if item != "Harvest Year"
                    ]
                )
                y_pred = self.model.predict(X_test)
                best_hyperparameters = {}
            else:
                y_pred = self.model.predict(X_test)
                best_hyperparameters = self.model.get_params().copy()

        if self.check_yield_trend:
            # Get information for retrending
            for idx, region in enumerate(df_region["Region"].unique()):
                mask_region = self.df_train["Region"] == df_region["Region"].unique()[0]
                df_tmp = self.df_train[mask_region]

                obj_trend = trend.DetrendedData(
                    df_tmp[f"Detrended {self.target}"],
                    df_tmp["Detrended Model"],
                    df_tmp["Detrended Model Type"],
                )

                # Retrend the predicted yield
                y_pred[idx] += trend.compute_trend(
                    obj_trend, df_region.iloc[idx][["Harvest Year"]]
                )[0]

                df_region.loc[
                    idx, "Detrended Model Type"
                ] = obj_trend.model_type.unique()[0]

        # Create a dataframe with forecast results
        shp = len(X_test)
        experiment_id = f"{self.country}_{self.crop}"
        now = ar.utcnow().to("America/New_York").format("MMMM-DD-YYYY HH:mm:ss")
        selected_features = self.selected_features + self.cat_features
        # Compute percentage difference between y_pred and y_test
        if self.model_type == "REGRESSION":
            ape = np.abs((y_pred - y_test) / y_test) * 100
        elif self.model_type == "CLASSIFICATION":
            # Create numpy array full of naNs with the same shape as y_test
            ape = np.full(shp, np.nan)

        df = pd.DataFrame(
            {
                "Experiment_ID": np.full(shp, experiment_id),
                "Experiment Name": np.full(shp, self.experiment_name),
                "Date": np.full(shp, self.today),
                "Time": np.full(shp, now),
                "Country": np.full(shp, self.country),
                "Crop": np.full(shp, self.crop),
                "Cluster Strategy": np.full(shp, self.cluster_strategy),
                "Frequency": np.full(shp, self.method),
                "Selected Features": [selected_features.copy() for _ in range(shp)],
                "Best Hyperparameters": np.full(shp, best_hyperparameters),
                "Stage_ID": np.full(shp, self.stage_info["Stage_ID"]),
                "Stage Range": np.full(shp, self.stage_info["Stage Range"]),
                "Stage Name": np.full(shp, self.stage_info["Stage Name"]),
                "Starting Stage": np.full(shp, self.stage_info["Starting Stage"]),
                "Ending Stage": np.full(shp, self.stage_info["Ending Stage"]),
                "Model": np.full(shp, self.model_name),
                "Region_ID": df_region["Region_ID"].values,
                "Region": df_region["Region"].values,
                "Harvest Year": df_region["Harvest Year"].values,
                "Area (ha)": df_region["Area (ha)"].values,
                f"Observed {self.target}": np.around(y_test, 3).ravel(),
                f"Predicted {self.target}": np.around(y_pred, 3).ravel(),
                "APE": np.around(ape, 3).ravel(),
            }
        )

        # Add median yield to dataframe
        df.loc[:, f"Median {self.target}"] = np.around(
            df_region[f"Median {self.target}"].values, 3
        )

        if f"Median {self.target} (2014-2018)" in df_region.columns:
            df.loc[:, f"Median {self.target} (2014-2018)"] = np.around(
                df_region[f"Median {self.target} (2014-2018)"].values, 3
            )

        if f"Median {self.target} (2013-2017)" in df_region.columns:
            df.loc[:, f"Median {self.target} (2013-2017)"] = np.around(
                df_region[f"Median {self.target} (2013-2017)"].values, 3
            )

        if self.estimate_ci:
            if self.estimate_ci_for_all or self.forecast_season == self.today_year:
                # Iterate over each element in y_pred_ci
                for idx, ci in enumerate(y_pred_ci):
                    df.loc[idx, "alpha"] = self.alpha
                    if self.model_type == "REGRESSION":
                        # Flatten the list
                        y_pred_ci_ = [item for sublist in ci for item in sublist]

                        df.loc[idx, f"lower CI"] = np.around(y_pred_ci_[0], 3)
                        df.loc[idx, f"upper CI"] = np.around(y_pred_ci_[1], 3)
                    elif self.model_type == "CLASSIFICATION":
                        df.loc[idx, f"CI"] = ", ".join(map(str, ci.flatten()))

        if self.check_yield_trend:
            df.loc[:, "Detrended Model Type"] = df_region["Detrended Model Type"].values

        if self.last_year_yield_as_feature:
            # Add last year yield to dataframe
            df.loc[:, f"Last Year {self.target}"] = np.around(
                df_region[f"Last Year {self.target}"].values, 3
            )

        if self.analogous_year_yield_as_feature:
            try:
                # Add analogous year and yield to dataframe
                df.loc[:, "Analogous Year"] = df_region["Analogous Year"].values
                df.loc[:, "Analogous Year Yield"] = np.around(
                    df_region["Analogous Year Yield"].values, 3
                )
            except:
                breakpoint()

        # if self.spatial_autocorrelation:
        #     # Compute spatial autocorrelation
        #     df = sa.compute_spatial_autocorrelation(
        #         self.dg_country
        #     )

        for col in [
            f"Median {self.target}",
            "Analogous Year",
            "Analogous Year Yield",
            "Detrended Model Type",
            "Detrended Model",
        ]:
            if col not in df.columns:
                df.loc[:, col] = np.NaN

        # Create an index based on following columns
        index_columns = [
            "Experiment Name",
            "Model",
            "Cluster Strategy",
            "Country",
            "Region",
            "Crop",
            "Harvest Year",
            "Stage Name",
            "Time",
        ]

        df.index = df.apply(
            lambda row: "_".join([str(row[col]) for col in index_columns]), axis=1
        )

        # name the index level
        df.index.set_names(["Index"], inplace=True)

        return experiment_id, df

    def create_feature_names(self, stages_features, selected_features):
        """
        Create feature names for machine learning stages.

        Args:
            stages_features (list): List of features for different stages.
            selected_features (dict): Dictionary of selected features.

        Returns:
            None
        """
        # Assert stages_features is a list
        assert isinstance(stages_features, list), "stages_features should be a list"

        # Clear out feature names
        self.feature_names = []

        """
        Select stages that will be used for ML
         1. method = "latest" - Select the latest stage
         2. method = "fraction" - Select a fraction (1-100) of all stages
        """
        method = "fraction"
        if self.model_name in ["cumulative_1", "cumulative_2", "cumulative_3"]:
            method = "latest"

        stages_features = stages.select_stages_for_ml(
            stages_features, method=method, n=60
        )

        for stage in stages_features:
            # Convert each element of stage to str and join with _
            _stage = "_".join(map(str, stage))

            # Create a list appending _stage to each element of combined_keys
            _tmp = [f"{col}_{_stage}" for col in self.combined_keys]

            for _t in _tmp:
                parts = _t.split("_")
                cei = parts[0] if parts[1].isdigit() else "_".join(parts[:2])

                try:
                    if self.model_name in [
                        "cumulative_1",
                        "cumulative_2",
                        "cumulative_3",
                    ]:
                        dict_fn = stages.get_stage_information_dict(_t, self.method)
                        tmp_col = f"{dict_fn['CEI']}"

                        if tmp_col in self.df_train.columns:
                            self.feature_names.append(tmp_col)
                    else:
                        # Check if any element of dict_selected_features is in _t
                        if selected_features["CEI"].any():
                            for x in selected_features["CEI"].values:
                                if x not in cei:
                                    continue

                                dict_fn = stages.get_stage_information_dict(
                                    _t, self.method
                                )
                                tmp_col = f"{dict_fn['CEI']} {dict_fn['Stage Name']}"

                                if tmp_col in self.df_train.columns:
                                    self.feature_names.append(tmp_col)
                except:
                    breakpoint()
        self.feature_names = list(set(self.feature_names))

        if self.median_yield_as_feature:
            self.feature_names.append(f"Median {self.target}")

        if self.lag_yield_as_feature:
            # For the number of years specified in self.number_lag_years
            for i in range(1, self.number_lag_years + 1):
                self.feature_names.append(f"t -{i} {self.target}")

        if self.analogous_year_yield_as_feature:
            self.feature_names.extend(["Analogous Year", "Analogous Year Yield"])

        if self.use_outlook_as_feature:
            self.feature_names.append("FCST")

        # Add lat and lon to feature names
        if self.include_lat_lon_as_feature:
            self.feature_names.extend(["lat", "lon"])

        self.selected_features = []

    def loop_ml(self, stages, dict_selected_features, dict_best_cei):
        """

        Args:
            stages:
            dict_selected_features:
            dict_best_cei:

        Returns:

        """
        from sklearn.preprocessing import StandardScaler

        scaler = StandardScaler() if self.model_name in ["linear", "gam"] else None

        """ Train, Predict, Explain and Store results for each region """
        pbar = tqdm(self.df_train["Region_ID"].unique(), leave=False)
        for idx, region in enumerate(pbar):
            if self.model_name in ["linear", "gam"]:
                self.create_feature_names(stages, dict_best_cei[region][0:3].tolist())
            elif self.model_name in ["cumulative_1", "cumulative_2", "cumulative_3"]:
                self.create_feature_names(stages, {})
            elif self.ml_model:
                self.create_feature_names(stages, dict_selected_features[region])
            elif self.model_name in ["median"]:
                self.feature_names = [f"Median {self.target}"]
                self.last_year_yield_as_feature = False
                self.analogous_year_yield_as_feature = False
            elif self.model_name in ["analog"]:
                self.feature_names = [f"Analogous Year", f"Analogous Year Yield"]
                self.last_year_yield_as_feature = False
                self.median_yield_as_feature = False

            mask_train = self.df_train["Region_ID"] == region
            mask_test = self.df_test["Region_ID"] == region

            num_regions_in_cluster = self.df_train[mask_train]["Region"].unique()

            if self.cluster_strategy == "individual":
                region_name = self.df_train["Region"].unique()[idx]
                pbar.set_description(f"Fit/Predict for {region_name}")
                pbar.update()
            elif self.cluster_strategy in ["auto_detect", "single"]:
                pbar.set_description(f"Fit/Predict for group {idx + 1}")
                pbar.update()

            common_columns = (
                [self.target, self.target_class]
                + self.statistics_columns
                + self.feature_names
                + [f"Median {self.target}"]
                + [f"Median {self.target} (2014-2018)"]
                + [f"Median {self.target} (2013-2017)"]
                + ["Region_ID"]
            )
            if self.check_yield_trend:
                common_columns += [
                    f"Detrended {self.target}",
                    "Detrended Model Type",
                    "Detrended Model",
                ]

            if self.last_year_yield_as_feature:
                common_columns += [f"Last Year {self.target}"]

            """ Train """
            # Filter dataframe based on region and self.feature_names
            df_region_train = self.df_train[mask_train]
            df_region_train = df_region_train[self.fixed_columns + common_columns]
            df_region_train.reset_index(drop=True, inplace=True)
            self.train(df_region_train, scaler)

            """ Predict """
            if self.check_yield_trend:
                # Exclude [f"Detrended {self.target}", "Detrended Model Type", "Detrended Model"]
                common_columns = common_columns[:-3]
            df_region_test = self.df_test[mask_test]
            df_region_test = df_region_test[self.fixed_columns + common_columns]
            df_region_test.reset_index(drop=True, inplace=True)
            experiment_id, df = self.predict(df_region_test, scaler)
            # df.reset_index(inplace=True)

            """ XAI """
            if self.do_xai:
                assert not self.estimate_ci, "Cannot perform XAI if estimate_ci is True"

                kwargs = {
                    "cluster_strategy": self.cluster_strategy,
                    "model": self.model,
                    "model_name": self.model_name,
                    "forecast_season": self.forecast_season,
                    "crop": self.crop,
                    "country": self.country,
                    "analysis_dir": self.dir_analysis,
                }
                df_xai = xai.explain(df_region_train, df_region_test, **kwargs)
                # df.set_index("Index", inplace=True)

            """ Output results to database """
            if not self.ml_model:
                model = self.model_name
            elif self.estimate_ci:
                try:
                    model = self.model.estimator_
                except:
                    if self.model_name == "catboost":
                        model = self.model
                    else:
                        model = self.model.estimator
            else:
                model = self.model

            try:
                output.store(self.db_path, experiment_id, df, model, self.model_name)
                wandb.log({experiment_id: df})
            except Exception as e:
                self.logger.error(f"Error storing results for {experiment_id} {e}")

    def get_cei_column_names(self, df):
        all_cei_columns = [
            col
            for col in df.columns
            if col not in self.fixed_columns + [self.target] + self.statistics_columns
        ]

        return all_cei_columns

    def create_ml_dataframe(self, df):
        """
        Create ML ready dataframe
        :param df:
        """
        _str = f"{self.country} {self.crop}"
        self.logger.info(f"Creating ML dataframe {_str}")

        # Convert from long to wide format
        df = df[
            ["Index", "Stage_ID", "CEI"]
            + self.fixed_columns
            + [self.target]
            + self.statistics_columns
        ]

        # For each combination of Index and Stage_ID, new columns have been
        # created and filled with the corresponding CEI value from the original dataframe
        # The new column names are a combination of the original Index and Stage_ID values
        # Use pivot_table to create the new columns and flatten the multi-index columns
        # HACK: Set missing values in self.target_col and self.statistics_columns to -1
        df.loc[:, [self.target] + self.statistics_columns] = df[
            [self.target] + self.statistics_columns
        ].fillna(-1)

        # HACK, replace later
        df.loc[:, "Area"] = df["Area"].fillna(-1)
        df = df.pivot_table(
            index=self.fixed_columns + [self.target] + self.statistics_columns,
            columns=["Index", "Stage_ID"],
            values="CEI",
        ).reset_index()

        # Reset self.target_col and self.statistics_columns to NaN
        df[[self.target] + self.statistics_columns] = df[
            [self.target] + self.statistics_columns
        ].replace(-1, np.nan)

        # Flatten the multi-index columns
        df.columns = [f"{i}_{j}" if j != "" else f"{i}" for i, j in df.columns]

        # Get all the columns apart from the fixed columns, target column and stats columns
        all_cei_columns = self.get_cei_column_names(df)
        parts = all_cei_columns[-1].split("_")
        cei = parts[0] if parts[1].isdigit() else "_".join(parts[:2])

        # For each region, find the column with the longest string in cei_column
        group_by = ["Region"]
        groups = df.groupby(group_by)
        if self.use_cumulative_features:
            frames = []
            for name, group in groups:
                # Drop columns with all NaNs
                group.dropna(axis=1, how="all", inplace=True)

                cei_column = group[
                    group.columns[group.columns.str.contains(cei)]
                ].columns
                max_cei_col = max(cei_column, key=len)
                self.stage_info = stages.get_stage_information_dict(
                    max_cei_col, self.method
                )

                # Subset dataframes to columns that contain self.stage_info["Stage_ID"]
                all_columns = group.columns[
                    group.columns.str.contains(self.stage_info["Stage_ID"])
                ].tolist()

                try:
                    group = group[
                        self.fixed_columns
                        + [self.target]
                        + self.statistics_columns
                        + all_columns
                    ]
                except:
                    continue
                # rename all_columns to self.stage_info["CEI"]
                group.rename(
                    columns={
                        col: stages.get_stage_information_dict(col, self.method)["CEI"]
                        for col in all_columns
                    },
                    inplace=True,
                )

                frames.append(group)

            df = pd.concat(frames)
        else:
            # HACK: Get feature name with GD4 in it to extract first and last stage id and name
            cei_column = df[df.columns[df.columns.str.contains(cei)]].columns
            # Select the longest string in cei_column
            cei_col = max(cei_column, key=len)
            self.stage_info = stages.get_stage_information_dict(cei_col, self.method)

        """ Only select those features that span a single time-period """
        """ e.g. vDTR_7_6 is ok but vDTR_7_6_5 is not """
        if self.use_single_time_period_as_feature:
            df = stages.select_single_time_period_features(df)

        # Change column name
        # e.g. 'vDTR_7_6_5_4_3_2_1_37_36_35_34_33_32_31' to 'vDTR Mar 1-Oct 27'
        df = stages.update_feature_names(df, self.method)

        all_cei_columns = self.get_cei_column_names(df)
        # Fill in any missing values with 0
        df.loc[:, all_cei_columns].fillna(0, inplace=True)

        df = fe.compute_last_year_yield(df, self.target)

        df = fe.compute_median_statistics(
            df, self.all_seasons_with_yield, self.number_median_years, self.target
        )

        df = fe.compute_user_median_statistics(
            df, [2014, 2015, 2016, 2017, 2018]
        )

        df = fe.compute_user_median_statistics(
            df, [2013, 2014, 2015, 2016, 2017]
        )

        if self.median_area_as_feature:
            df = fe.compute_median_statistics(
                df, self.all_seasons_with_yield, self.number_median_years, "Area (ha)"
            )

        if self.lag_yield_as_feature:
            df = fe.compute_lag_yield(
                df, self.all_seasons_with_yield, self.number_lag_years, self.target
            )

        if self.analogous_year_yield_as_feature:
            df = fe.compute_analogous_yield(
                df, self.all_seasons_with_yield, self.number_median_years, self.target
            )

        # Create Region_ID column based on Region column category code
        df["Region"] = df["Region"].astype("category")
        if self.cluster_strategy == "single":
            df["Region_ID"] = 1
        elif self.cluster_strategy == "individual":
            df["Region_ID"] = df["Region"].cat.codes
        elif self.cluster_strategy == "auto_detect":
            clusters_assigned = fe.detect_clusters(df, self.target)
            # Merge the cluster labels with the original DataFrame
            df = df.merge(clusters_assigned, on="Region")

            # Region_ID should be type category
            df["Region_ID"] = df["Region_ID"].astype("category")

        return df

    def execute(self):
        """

        Returns:

        """
        """ Get dataframe including all simulation stages """
        # convert self.simulation_stages to a list of strings
        _stages = []
        for stage in self.simulation_stages:
            stage = stages.convert_stage_string(stage, to_array=False)
            _stages.append(stage)

        mask = self.df_inputs["Stage_ID"].isin(_stages)
        df = self.df_inputs[mask]
        """ Select which CEI categories to use for ML """
        if "all" in self.use_ceis:
            pass
        else:
            if self.select_cei_by == "Type":
                df = df[df["Type"].isin(self.use_ceis)]
            elif self.select_cei_by == "Index":
                df = df[df["Index"].isin(self.use_ceis)]

        """ Convert this dataframe into an ML ready format and save to disk """
        df = self.create_ml_dataframe(df)
        dir_output = (
            self.dir_analysis
            / self.country
            / self.crop
            / self.model_name
            / str(self.forecast_season)
        )
        os.makedirs(dir_output, exist_ok=True)
        df.to_csv(
            dir_output / f"{self.country}_{self.crop}_{self.forecast_season}.csv",
            index=False,
        )

        # cat_features should be converted to category type
        df[self.cat_features] = df[self.cat_features].astype("category")

        """  Heatmap of correlation of various features with yield at each time step"""
        # For each time step, determine the number of occurences of each feature
        # when it is most correlated with the target for a region
        # Use this to plot a heatmap showing which features are most correlated
        # with yield at each time step
        # Add a Country Region column to df and make it lower case
        df["Country Region"] = (
            df["Country"].astype(str) + " " + df["Region"].astype(str)
        ).str.lower()

        # Join with dg based on Country Region column, only keeping rows that are in df
        # Only use geometry column from self.dg
        self.dg_country = self.dg_country[["Country Region", "geometry"]].merge(
            df[["Country Region", self.correlation_plot_groupby]],
            on="Country Region",
            how="outer",
        )

        # Add a lat and lon column to self.dg_country
        self.dg_country["lat"] = self.dg_country.centroid.y
        self.dg_country["lon"] = self.dg_country.centroid.x

        # Add lat and lon columns to df by merging on Country Region column
        df = df.merge(
            self.dg_country[["Country Region", "lat", "lon"]].drop_duplicates(),
            on="Country Region",
            how="left",
        )

        dict_kwargs = {}
        dict_kwargs["all_stages"] = self.all_stages
        dict_kwargs["target_col"] = self.target
        dict_kwargs["country"] = self.country
        dict_kwargs["crop"] = self.crop
        dict_kwargs["dir_output"] = (
            self.dir_analysis
            / self.country
            / self.crop
            / self.model_name
            / str(self.forecast_season)
        )
        dict_kwargs["forecast_season"] = self.forecast_season
        dict_kwargs["method"] = self.method
        dict_kwargs["national_correlation"] = self.national_correlation
        dict_kwargs["groupby"] = self.correlation_plot_groupby
        dict_kwargs["cluster_strategy"] = self.cluster_strategy
        dict_kwargs["dg_country"] = self.dg_country
        dict_kwargs["combined_dict"] = self.combined_dict
        dict_kwargs["plot_map"] = self.plot_map_for_correlation_plot
        dict_kwargs["correlation_threshold"] = self.correlation_threshold

        if self.spatial_autocorrelation:
            sa.compute_spatial_autocorrelation(self.df_inputs, **dict_kwargs)

        dict_selected_features = {}
        dict_best_cei = {}
        if self.correlation_plots:
            self.logger.info(f"Correlation plot for {self.country} {self.crop}")
            (
                dict_selected_features,
                dict_best_cei,
            ) = correlations.all_correlated_feature_by_time(df, **dict_kwargs)

        """ Add Yield_class column and set to NaN """
        df[f"{self.target}_class"] = np.NaN

        """ Separate into train and test datasets based on forecast_season """
        mask = df["Harvest Year"] == self.forecast_season
        self.df_train = df[~mask]
        self.df_test = df[mask]

        # Drop rows with missing values for self.target_col in df_train
        self.df_train = self.df_train.dropna(subset=[self.target])

        """ Groupby Region column and compute detrended yield """
        self.df_train[f"Detrended {self.target}"] = np.NaN
        self.df_train["Detrended Model"] = np.NaN
        self.df_train["Detrended Model Type"] = np.NaN
        group_by = ["Region"]
        groups = self.df_train.groupby(group_by)
        for key, group in groups:
            if group.empty:
                continue

            if self.check_yield_trend:
                if group[self.target].any():
                    detrended_data = trend.detrend_dataframe(
                        group, column_name=self.target, model_type="linear"
                    )
                    self.df_train.loc[
                        group.index, f"Detrended {self.target}"
                    ] = detrended_data.detrended_series
                    self.df_train.loc[
                        group.index, "Detrended Model"
                    ] = detrended_data.trend_model
                    self.df_train.loc[
                        group.index, "Detrended Model Type"
                    ] = detrended_data.model_type

            # Create categorical classes for target column
            group, new_target_column, bins = fe.classify_target(
                group, self.target, self.number_classes
            )
            self.target_bins[key[0]] = bins
            self.target_class = new_target_column
            self.df_train.loc[group.index, new_target_column] = group[new_target_column]

        # 6. Exclude years without yields from df_train
        # self.df_train = self.df_train[
        #     self.df_train["Harvest Year"].isin(self.all_seasons_with_yield)
        # ]

        """ Run ML code for each stage """
        if self.run_ml:
            self.logger.info(f"Running ML for {self.country} {self.crop}")
            setup_stages = [self.simulation_stages]
            if self.run_latest_time_period:
                # setup_stages = stages.select_stages_for_ml(setup_stages, method="latest")
                setup_stages = [setup_stages[-1]]

            num_regions = len(self.df_train["Region_ID"].unique())
            pbar = tqdm(setup_stages)
            for stage in pbar:
                pbar.set_description(
                    f"ML {num_regions} regions, {len(setup_stages)} stages"
                )
                pbar.update()

                self.loop_ml(stage, dict_selected_features, dict_best_cei)
        wandb.finish()

    def setup(self, forecast_season, model):
        """

        :param country:
        :param crop:
        :param forecast_season:
        :param model:
        """
        _str = f"{self.country} {self.crop} {model} {forecast_season}"
        config_dict = {section: dict(self.parser.items(section)) for section in self.parser.sections()}
        wandb.init(project=self.project_name, name=_str, config=config_dict)
        self.logger.info(f"Setup {_str}")

        self.forecast_season = forecast_season
        self.model_name = model
        self.experiment_name = self.parser.get("ML", "experiment_name")
        self.ml_model = self.parser.getboolean(self.model_name, "ML_model")
        self.select_cei_by = self.parser.get(self.model_name, "select_cei_by")
        self.use_ceis = ast.literal_eval(self.parser.get(self.model_name, "use_ceis"))
        self.model_names = ast.literal_eval(self.parser.get(self.country, "models"))
        self.optimize = self.parser.getboolean(self.country, "optimize")
        self.fraction_loocv = self.parser.getfloat(self.country, "fraction_loocv")
        self.all_seasons = self.df_inputs["Harvest Year"].unique()

        # If model_type if regression then set classify_target to False
        if self.model_type == "REGRESSION" and self.classify_target:
            raise ValueError("Model type is regression but classify_target is True")
        elif self.model_type == "CLASSIFICATION" and not self.classify_target:
            raise ValueError(
                "Model type is classification but classify_target is False"
            )

        """ For a classification model, only catboost is supported currently """
        if self.model_type == "CLASSIFICATION":
            self.do_xai = False
            self.alpha = self.parser.getfloat("ML", "alpha")
            self.estimate_ci = self.parser.getboolean("ML", "estimate_ci")
            self.estimate_ci_for_all = self.parser.getboolean(
                "ML", "estimate_ci_for_all"
            )
            self.check_yield_trend = False
            if self.model_name == "ngboost":
                # Remove Region from cat_features as it is object type
                self.cat_features = [
                    col for col in self.cat_features if col != "Region"
                ]
        elif self.model_type == "REGRESSION":
            """If not using a ML model then set XAI and CI to False"""
            if not self.ml_model or self.model_name in ["linear", "gam", "merf"]:
                self.do_xai = False
                self.estimate_ci = False
                self.check_yield_trend = False
                self.estimate_ci_for_all = False
            elif self.model_name in ["cumulative_1", "cumulative_2", "cumulative_3"]:
                self.correlation_plots = False
                self.lag_yield_as_feature = False
                self.median_yield_as_feature = False
                self.median_area_as_feature = False
                self.analogous_year_yield_as_feature = False
                self.last_year_yield_as_feature = False
                self.include_lat_lon_as_feature = False
                self.do_xai = False
                self.estimate_ci = False
                self.estimate_ci_for_all = False
                self.check_yield_trend = True
                self.cluster_strategy = "single"
                self.select_cei_by = "Index"
                self.use_cumulative_features = True
            elif self.model_name in ["tabpfn"]:
                self.do_xai = False
                self.estimate_ci = False
            elif self.model_name in ["oblique", "ydf"]:
                self.do_xai = False
                self.estimate_ci = False
                # Remove Region from cat_features as it is object type
                self.cat_features = [
                    col for col in self.cat_features if col != "Region"
                ]
            elif self.model_name == "ngboost":
                self.do_xai = False
                self.alpha = self.parser.getfloat("ML", "alpha")
                self.estimate_ci = self.parser.getboolean("ML", "estimate_ci")
                self.estimate_ci_for_all = self.parser.getboolean(
                    "ML", "estimate_ci_for_all"
                )
                # Remove Region from cat_features as it is object type
                self.cat_features = [
                    col for col in self.cat_features if col != "Region"
                ]
            else:
                self.do_xai = self.parser.getboolean("ML", "do_xai")
                self.estimate_ci = self.parser.getboolean("ML", "estimate_ci")
                self.estimate_ci_for_all = self.parser.getboolean(
                    "ML", "estimate_ci_for_all"
                )
                self.alpha = self.parser.getfloat("ML", "alpha")
                self.check_yield_trend = self.parser.getboolean(
                    "ML", "check_yield_trend"
                )

        # 1. Get all seasons with yield, these are the seasons for which Yield is available
        self.all_seasons_with_yield = self.df_inputs[
            self.df_inputs[self.target].notna()
        ]["Harvest Year"].unique()

        if self.method.endswith("_r"):
            if self.forecast_season == self.today_year:
                mask = self.df_inputs["Harvest Year"] == self.forecast_season
                self.all_stages = self.df_inputs[mask]["Stage_ID"].unique()
            else:
                self.all_stages = self.df_inputs["Stage_ID"].unique()
        else:
            raise NotImplementedError(f"Method {self.method} not implemented")

        # Create a list of numpy arrays, for each string in _stages,
        # create a numpy array by splitting by _ and converting to int
        # then store in self.simulation_stages
        # e.g. _stages = ['13_12_11', '13_12_11_10', '13_12_11_10_9']
        # then self.simulation_stages = [array([13, 12, 11]), array([13, 12, 11, 10]), array([13, 12, 11, 10, 9])]
        # Drop stages in self.all_stages that do not have _ in them
        self.all_stages = [element for element in self.all_stages if "_" in element]

        self.simulation_stages = [
            np.array([int(stage) for stage in s.split("_")]) for s in self.all_stages
        ]

        self.name_shapefile = self.parser.get(self.country, "boundary_file")
        self.admin_zone = self.parser.get(self.country, "admin_zone")
        self.dg = gp.read_file(
            self.dir_shapefiles / self.name_shapefile,
            engine="pyogrio",
        )

        if "ADMIN0" in self.dg.columns:
            # Hack rename Tanzania to United Republic of Tanzania
            self.dg["ADMIN0"] = self.dg["ADMIN0"].replace(
                "Tanzania", "United Republic of Tanzania"
            )

        # Rename ADMIN0 to ADM0_NAME and ADMIN1 to ADM1_NAME and ADMIN2 to ADM2_NAME
        self.dg = self.dg.rename(
            columns={
                "ADMIN0": "ADM0_NAME",
                "ADMIN1": "ADM1_NAME",
                "ADMIN2": "ADM2_NAME",
            }
        )

        if self.country == "nepal":
            self.dg["ADM0_NAME"] = "nepal"
            self.dg["Country Region"] = self.dg["ADM0_NAME"] + " " + self.dg["PR_NAME"]
        elif self.country == "wolayita":
            self.dg["ADM0_NAME"] = "ethiopia"
            self.dg["Country Region"] = self.dg["ADM0_NAME"] + " " + self.dg["W_NAME"]
        elif self.admin_zone == "admin_1":
            self.dg["Country Region"] = (
                self.dg["ADM0_NAME"] + " " + self.dg["ADM1_NAME"]
            )
        elif self.country == "illinois":
            self.dg["ADM0_NAME"] = "illinois"
            self.dg["Country Region"] = self.dg["ADM0_NAME"] + " " + self.dg["NAME"]
        else:
            self.dg["Country Region"] = (
                self.dg["ADM0_NAME"] + " " + self.dg["ADM2_NAME"]
            )
        self.dg["Country Region"] = self.dg["Country Region"].str.lower()
        self.dg_country = self.dg[
            self.dg["ADM0_NAME"].str.lower().str.replace(" ", "_") == self.country
        ]

    def read_data(self, country, crop, season):
        """

        Args:
            country:
            crop:
            season:

        Returns:

        """
        self.logger.info(f"Reading data for {country} {crop} {season}")

        self.country = country
        self.crop = crop

        """Figure out path to file containing CEI data"""
        admin_zone = self.parser.get(country, "admin_zone")
        country_str = country.title().replace("_", " ")
        crop_str = crop.title().replace("_", " ")

        dir_statistics = self.dir_output / "cei" / "indices" / self.method / "global"
        os.makedirs(dir_statistics, exist_ok=True)
        file = (
            dir_statistics / f"{country_str}_{crop_str}_statistics_s1_{self.method}.csv"
        )

        """ Add area, yield and prod stats from either GEOGLAM or FEWSNET warehouse database"""
        if not os.path.exists(file) or self.update_input_file:
            _dir_country = (
                self.dir_output
                / "cei"
                / "indices"
                / self.method
                / admin_zone
                / country
                / crop
            )
            file_name = f"{country}_{crop}_s1*.csv"
            all_files = _dir_country.glob(file_name)
            # TODO ignore file with _2000 in its name
            all_files = [f for f in all_files if "_2000" not in f.name]

            # Assert that all_files is not empty
            assert all_files, f"No files found in {_dir_country} with {file_name}"

            self.df_inputs = pd.concat(
                (pd.read_csv(f) for f in all_files), ignore_index=True
            )

            self.df_inputs = stats.add_statistics(
                self.dir_condition / "yield",
                self.df_inputs,
                country_str,
                crop_str,
                admin_zone,
                [self.target] + self.statistics_columns,
                self.method,
            )
            """ Add information on starting and ending time period for each stage"""
            self.df_inputs = stages.add_stage_information(self.df_inputs, self.method)

            self.df_inputs.to_csv(file, index=False)
        else:
            self.df_inputs = pd.read_csv(file)

        # Rename target column
        if self.rename_target:
            self.df_inputs.rename(
                columns={self.target: self.new_name_target}, inplace=True
            )
            self.target = "Yield"



import os

import numpy as np
import pandas as pd
from tqdm import tqdm


def compute_last_year_yield(df, target_col="Yield (tn per ha)"):
    """
    Computes the yield of the previous year for each region.

    Args:
        df (DataFrame): The original DataFrame containing yield data.
        target_col (str): The column name from which to compute the previous year yield.

    Returns:
        DataFrame: The original DataFrame enhanced with a new column for the previous year yield.
    """
    # Ensure 'Harvest Year' is treated as integer for accurate comparisons
    df["Harvest Year"] = df["Harvest Year"].astype(int)
    # Initialize the new column with NaNs
    df[f"Last Year {target_col}"] = np.nan

    for region, group in tqdm(
        df.groupby("Region"), desc="Last year yields", leave=False
    ):
        unique_years = group["Harvest Year"].unique()

        for harvest_year in unique_years:
            mask = (group["Harvest Year"] == harvest_year - 1) & (
                group["Region"] == region
            )
            last_year_yield = group.loc[mask, target_col].values
            if last_year_yield:
                df.loc[
                    (df["Region"] == region) & (df["Harvest Year"] == harvest_year),
                    f"Last Year {target_col}",
                ] = last_year_yield[0]

    return df


def compute_closest_years(all_years, harvest_year, number_lag_years):
    """
    Finds the years closest to a given harvest year, excluding the harvest year itself.

    Args:
        all_years (array-like): List or array of all years to consider.
        harvest_year (int): The year from which to compute distance.
        number_lag_years (int): Number of closest years to return.

    Returns:
        list: Years closest to the given harvest year.
    """
    # Exclude the harvest year before computation to simplify logic
    filtered_years = [year for year in all_years if year != harvest_year]

    closest_years = np.array(filtered_years)[
        np.argsort(np.abs(np.array(filtered_years) - harvest_year))[:number_lag_years]
    ]

    return closest_years.tolist()


def compute_median_statistics(
    df, all_seasons_with_yield, number_median_years, target_col="Yield (tn per ha)"
):
    """
    Enhances the DataFrame with a new column that contains the median yield from the closest lag years.

    Args:
        df (DataFrame): The original DataFrame containing yield data.
        all_seasons_with_yield (array-like): List of seasons that have yield data.
        number_median_years (int): Number of years to consider for computing the median yield.
        target_col (str): The column name from which to compute the median yield.

    Returns:
        DataFrame: The original DataFrame enhanced with a new column for median lag yield.
    """
    # Ensure 'Harvest Year' is treated as integer for accurate comparisons
    df["Harvest Year"] = df["Harvest Year"].astype(int)
    # Initialize the new column with NaNs
    df[f"Median {target_col}"] = np.nan

    for region, group in tqdm(df.groupby("Region"), desc="Median yield", leave=False):
        unique_years = group["Harvest Year"].unique()

        # Check if the target column is empty for the current group
        if group[target_col].isnull().all():
            continue

        for harvest_year in unique_years:
            closest_years = compute_closest_years(
                all_seasons_with_yield, harvest_year, number_median_years
            )
            mask = (group["Harvest Year"].isin(closest_years)) & (
                group["Region"] == region
            )
            median_yield = group.loc[mask, target_col].median()
            df.loc[
                (df["Region"] == region) & (df["Harvest Year"] == harvest_year),
                f"Median {target_col}",
            ] = median_yield

    return df


def compute_user_median_statistics(df, user_years, target_col="Yield (tn per ha)"):
    """
    Enhances the DataFrame with a new column that contains the median yield computed
    using only the yields from the user-specified list of years.

    Args:
        df (DataFrame): The original DataFrame containing yield data.
        user_years (array-like): List of years to consider for computing the median yield.
        target_col (str): The column name from which to compute the median yield.

    Returns:
        DataFrame: The original DataFrame enhanced with a new column for median yield.
    """
    # Ensure 'Harvest Year' is treated as integer for accurate comparisons.
    df["Harvest Year"] = df["Harvest Year"].astype(int)

    # Sort the user_years list to reliably extract the earliest and latest years.
    user_years_sorted = sorted(user_years)
    first_year = user_years_sorted[0]
    last_year = user_years_sorted[-1]

    # Define the new column name to include the range of years.
    new_col_name = f"Median {target_col} ({first_year}-{last_year})"

    # Initialize the new column with NaN values.
    df[new_col_name] = np.nan

    # Group by region and compute the median yield for the specified years.
    for region, group in tqdm(df.groupby("Region"), desc="Median yield", leave=False):
        # Skip if the target column is completely null for this region.
        if group[target_col].isnull().all():
            continue

        # Filter the rows to only include harvest years that are in the user provided list.
        mask = group["Harvest Year"].isin(user_years)
        median_yield = group.loc[mask, target_col].median()

        # Assign the computed median yield to all rows in the current region.
        df.loc[df["Region"] == region, new_col_name] = median_yield

    return df


def compute_lag_yield(
    df, all_seasons_with_yield, number_lag_years, target_col="Yield (tn per ha)"
):
    # For the number of years specified in self.number_lag_years, add the yield of that number of years
    # ago to the dataframe
    # For example, if number_lag_years is 3, then the yield of each year upto 3 years ago will be added
    # to the dataframe
    # The yield of the previous year is already added to the dataframe
    # Ensure 'Harvest Year' is treated as integer for accurate comparisons
    df["Harvest Year"] = df["Harvest Year"].astype(int)

    for region, group in tqdm(df.groupby("Region"), desc="Lag yields", leave=False):
        unique_years = group["Harvest Year"].unique()

        # Check if the target column is empty for the current group
        if group[target_col].isnull().all():
            continue

        for harvest_year in unique_years:
            closest_years = compute_closest_years(
                all_seasons_with_yield, harvest_year, number_lag_years
            )

            # For each year in the closest years, add the yield to the dataframe as a new column
            for idx, year in enumerate(closest_years):
                col = f"t -{idx + 1} {target_col}"

                mask_group_year = group["Harvest Year"] == year
                mask_region = (df["Region"] == region) & (
                    df["Harvest Year"] == harvest_year
                )
                yield_value = group.loc[mask_group_year, target_col].values

                if yield_value.size > 0:
                    df.loc[mask_region, col] = yield_value[0]
                else:
                    # Add median yield
                    mask_group_median = group["Harvest Year"].isin(closest_years)
                    median_yield = group.loc[mask_group_median, target_col].median()

                    df.loc[mask_region, col] = median_yield

    return df


def compute_analogous_yield(
    df,
    all_seasons_with_yield,
    number_lag_years,
    target_col="Yield (tn per ha)",
    var="ESI4WK",
):
    """
    Computes and adds analogous year and its yield based on the similarity of environmental conditions.

    Args:
        df (pd.DataFrame): Input dataframe with yield and environmental data.
        all_seasons_with_yield (array-like): List of seasons that have yield data.
        number_lag_years (int): Number of years to consider for finding analogous years.
        target_col (str): The column name to use for yield data.
        var (str): The environmental variable prefix to find similarity.

    Returns:
        pd.DataFrame: The dataframe with added columns for analogous year and its yield.
    """
    from sklearn.metrics import root_mean_squared_error

    # Determine relevant columns based on the environmental variable
    if "ESI4WK" in var:
        var_columns = [col for col in df.columns if var in col]
    else:
        var_columns = [col for col in df.columns if "ESI" not in col]

    # Early exit if only one variable column is found
    if len(var_columns) == 1:
        df["Analogous Year"] = np.nan
        df["Analogous Year Yield"] = df[f"Median {target_col}"]
        return df

    # Initialize the new columns to NaN
    df["Analogous Year"] = np.nan
    df["Analogous Year Yield"] = np.nan

    all_years = df["Harvest Year"].unique()

    for harvest_year in tqdm(all_years, desc="Computing analogous yields", leave=False):
        lag_years = compute_closest_years(
            all_seasons_with_yield, harvest_year, number_lag_years
        )

        for region in df["Region"].unique():
            # Filter current year and region dataset
            df_current = df[
                (df["Harvest Year"] == harvest_year) & (df["Region"] == region)
            ]
            # Filter dataset for lag years and the same region
            df_lag = df[(df["Harvest Year"].isin(lag_years)) & (df["Region"] == region)]

            if df_current.empty or df_lag.empty:
                continue  # Skip if no data available for comparison

            # Calculate RMSE between the current year's profile and each of the lag years' profiles
            min_rmse, analogous_year, analogous_yield = np.inf, np.nan, np.nan
            for _, row_current in df_current.iterrows():
                for _, row_lag in df_lag.iterrows():
                    # Remove NaNs from both row_current and row_lag
                    arr1 = row_current[var_columns]
                    arr2 = row_lag[var_columns]

                    # Identify the positions where array1 is not NaN
                    not_nan_indices = ~np.isnan(arr1.astype("float").values)

                    # Remove NaNs from array1 and corresponding elements from array2
                    arr1 = arr1[not_nan_indices]
                    arr2 = arr2[not_nan_indices]

                    try:
                        rmse = root_mean_squared_error(arr1, arr2)
                    except:
                        continue
                    if rmse < min_rmse:
                        min_rmse = rmse
                        analogous_year = row_lag["Harvest Year"]
                        analogous_yield = row_lag[target_col]

            # Update the DataFrame with the found analogous year and yield
            mask = (df["Region"] == region) & (df["Harvest Year"] == harvest_year)
            df.loc[mask, "Analogous Year"] = analogous_year
            df.loc[mask, "Analogous Year Yield"] = (
                analogous_yield
                if not np.isnan(analogous_yield)
                else df.loc[mask, f"Median {target_col}"]
            )

    return df


def detect_clusters(df, target_col="Yield (tn per ha)"):
    """

    Args:
        df:
        target_col:

    Returns:

    """
    os.environ["OMP_NUM_THREADS"] = "1"

    # Suppress warnings in this function
    import warnings

    warnings.filterwarnings("ignore")

    from kneed import KneeLocator
    from sklearn.cluster import KMeans

    # Get the yield of each region in df_results
    df_yield = df[["Region", "Harvest Year", target_col]]

    # Drop any columns with missing values
    df_yield = df_yield.dropna()

    # Pivot the DataFrame to have regions as rows and years as columns with their corresponding yields
    df_yield_pivot = df_yield.pivot_table(
        index="Region",
        columns="Harvest Year",
        values=target_col,
        aggfunc="mean",
    )

    # Fill rows with median value of the row
    df_na = df_yield_pivot.median(axis=1)
    # Fill NaNs in df_yield_pivot with value from df_na
    df_yield_pivot = df_yield_pivot.fillna(df_na, axis=0)

    # Fill data with Median values rowwise
    df_yield_pivot = df_yield_pivot.apply(lambda row: row.fillna(row.median()), axis=1)

    # Normalize the data
    # scaler = StandardScaler()
    # df_yield_normalized = scaler.fit_transform(df_yield_pivot)

    # Determine an appropriate number of clusters using the Elbow Method
    inertia = []
    range_of_clusters = range(
        1, len(df_yield_pivot)
    )  # Assuming up to 15 clusters for demonstration

    # Replace inf or NaN values with column median
    df_yield_pivot = df_yield_pivot.replace([np.inf, -np.inf], np.nan)
    df_yield_pivot = df_yield_pivot.fillna(df_yield_pivot.median())
    for n_clusters in range_of_clusters:
        kmeans = KMeans(n_clusters=n_clusters, random_state=42)
        try:
            kmeans.fit(df_yield_pivot)
        except:
            breakpoint()
        inertia.append(kmeans.inertia_)

    # Use KneeLocator to find the elbow point automatically
    knee_locator = KneeLocator(
        range_of_clusters, inertia, curve="convex", direction="decreasing"
    )

    # # Plot the Elbow Method for visual confirmation
    # plt.figure(figsize=(10, 6))
    # plt.plot(range_of_clusters, inertia, marker='o', linestyle='--')
    # plt.title('Elbow Method For Optimal k')
    # plt.xlabel('Number of clusters')
    # plt.ylabel('Inertia')
    # plt.xticks(range_of_clusters)
    # plt.vlines(knee_locator.knee, plt.ylim()[0], plt.ylim()[1], linestyles='dashed')
    # plt.show()

    # Use the detected number of clusters
    optimal_clusters = knee_locator.knee
    if optimal_clusters:
        optimal_clusters = (
            optimal_clusters + 1 if optimal_clusters > 1 else optimal_clusters
        )

        # Apply K-Means clustering with the detected optimal number of clusters
        kmeans = KMeans(n_clusters=optimal_clusters, random_state=42)
        kmeans.fit(df_yield_pivot)

        # Assign cluster labels to each region
        cluster_labels = kmeans.labels_

        # Create a DataFrame with region names and their respective cluster IDs
        clusters_assigned = pd.DataFrame(
            {"Region": df_yield_pivot.index, "Region_ID": cluster_labels}
        )
    else:
        # If no optimal_clusters is found, then assign all regions to a single cluster
        clusters_assigned = pd.DataFrame(
            {"Region": df_yield_pivot.index, "Region_ID": 1}
        )

    return clusters_assigned


def classify_target(df, target_col, number_classes):
    """

    Args:
        df:
        target_col:
        number_classes:

    Returns:

    """
    new_target_col = f"{target_col}_class"

    # Change the target column to categorical with the specified number of classes
    df[new_target_col], bins = pd.qcut(df[target_col],
                                       q=number_classes,
                                       labels=False,
                                       retbins=True,
                                       duplicates='drop')

    return df, new_target_col, bins


# breakpoint()

# from libpysal.weights import Queen, Rook
# from pysal.lib import weights
# from scipy.linalg import eigh
#
# breakpoint()
# df = pd.DataFrame()
#
# # Create a spatial weights matrix (e.g., Queen contiguity)
# w = weights.Queen.from_dataframe(dg)
#
# # Transform weights to row-standardized form
# w.transform = 'r'
#
# # Convert the weights matrix to a dense format for eigen decomposition
# W_dense = w.full()[0]
#
# # Compute eigenvalues and eigenvectors
# eigenvalues, eigenvectors = eigh(W_dense)
#
# # Sort eigenvalues and corresponding eigenvectors
# sorted_indices = np.argsort(eigenvalues)[::-1]
# eigenvalues = eigenvalues[sorted_indices]
# eigenvectors = eigenvectors[:, sorted_indices]
#
# # Select a subset of eigenvectors (e.g., first 10)
# selected_eigenvectors = eigenvectors[:, :2]
#
# breakpoint()
# # Add eigenvectors to the GeoDataFrame
# for i in range(selected_eigenvectors.shape[1]):
#     df[f'EV_{i + 1}'] = selected_eigenvectors[:, i]



import os
import ee
import geemap
import pandas as pd
import geopandas as gpd
from datetime import datetime
import logging

# Configure Logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s'
)

# Initialize Earth Engine
try:
    ee.Initialize()
    logging.info("Successfully initialized Earth Engine.")
except Exception as e:
    logging.info("Initializing Earth Engine Authentication.")
    ee.Authenticate()
    ee.Initialize()
    logging.info("Successfully authenticated and initialized Earth Engine.")


class SatelliteDataProcessor:
    def __init__(self, shapefile_path, start_date_col, end_date_col, data_source='HLS',
                 output_folder='downloads', stats=['mean'], percentiles=[25, 50, 75],
                 field_id_col=None, buffer_size=30):
        """
        Initialize the SatelliteDataProcessor with user-supplied parameters.
        """
        self.shapefile_path = shapefile_path
        self.start_date_col = start_date_col
        self.end_date_col = end_date_col
        self.data_source = data_source
        self.output_folder = output_folder
        self.stats = stats
        self.percentiles = percentiles
        self.field_id_col = field_id_col
        self.buffer_size = buffer_size
        self.gdf = None
        self.ee_feature_collection = None

        # Ensure output directory exists
        os.makedirs(self.output_folder, exist_ok=True)
        logging.info(f"Output directory '{self.output_folder}' is ready.")

        # Load and prepare shapefile
        self.load_shapefile()

    def load_shapefile(self):
        """
        Load the shapefile into a GeoDataFrame, ensure valid geometries,
        handle CRS for accurate buffering, and convert to GEE FeatureCollection.
        """
        try:
            self.gdf = gpd.read_file(self.shapefile_path)
            logging.info(f"Shapefile '{self.shapefile_path}' loaded successfully.")
        except Exception as e:
            logging.error(f"Failed to load shapefile '{self.shapefile_path}': {e}")
            raise

        # Ensure geometries are valid
        self.gdf['geometry'] = self.gdf['geometry'].apply(
            lambda geom: geom if geom.is_valid else geom.buffer(0)
        )
        logging.info("Validated and corrected geometries.")

        # Handle CRS: Buffering requires a projected CRS. Reproject if necessary.
        if self.gdf.crs is None:
            logging.error("Shapefile has no CRS defined. Please define a CRS.")
            raise ValueError("Shapefile has no CRS defined.")
        else:
            logging.info(f"Shapefile CRS is projected: {self.gdf.crs}")

        # Convert date columns to proper format
        self.gdf[self.start_date_col] = self.gdf[self.start_date_col].apply(self.format_date)
        self.gdf[self.end_date_col] = self.gdf[self.end_date_col].apply(self.format_date)
        logging.info("Formatted date columns.")

        # Convert GeoDataFrame to GEE FeatureCollection
        self.ee_feature_collection = self.gdf_to_gee(self.gdf)
        logging.info("Converted GeoDataFrame to Earth Engine FeatureCollection.")

    @staticmethod
    def get_utm_crs(lon, lat):
        """
        Determine the UTM CRS based on longitude and latitude.
        """
        utm_band = int((lon + 180) / 6) + 1
        if lat >= 0:
            epsg_code = 32600 + utm_band
        else:
            epsg_code = 32700 + utm_band
        return f"EPSG:{epsg_code}"

    @staticmethod
    def format_date(date_input):
        """
        Convert date input to 'YYYY-MM-DD' format or 'YYYY' if only year is provided.
        Supports 'DD-MM-YYYY', 'YYYY-MM-DD', 'YYYY', and float formats.
        """
        if pd.isnull(date_input):
            return None
        # Handle float inputs
        if isinstance(date_input, float):
            if date_input.is_integer():
                date_input = int(date_input)
                logging.debug(f"Converted float {date_input} to integer for date processing.")
            else:
                logging.warning(f"Float date '{date_input}' has a fractional part. Treating as invalid date.")
                return None
        # Convert to string if not already
        if not isinstance(date_input, str):
            date_input = str(date_input).strip()
        else:
            date_input = date_input.strip()

        # Attempt to parse full dates
        for fmt in ('%d-%m-%Y', '%Y-%m-%d'):
            try:
                return datetime.strptime(date_input, fmt).strftime('%Y-%m-%d')
            except ValueError:
                continue
        # Attempt to parse year-only format
        try:
            return datetime.strptime(date_input, '%Y').strftime('%Y')
        except ValueError:
            pass
        logging.warning(
            f"Invalid date format: '{date_input}'. Expected 'DD-MM-YYYY', 'YYYY-MM-DD', 'YYYY', or integer year.")
        return None

    @staticmethod
    def gdf_to_gee(gdf):
        """
        Converts a GeoDataFrame to an Earth Engine FeatureCollection.
        """
        features = []
        for idx, row in gdf.iterrows():
            geom = row['geometry'].__geo_interface__
            # Use specified field identifier or 'id'
            properties = row.drop('geometry').to_dict()
            feature = ee.Feature(ee.Geometry(geom), properties)
            features.append(feature)
        return ee.FeatureCollection(features)

    def mask_edge(self, image):
        """
        Apply edge masking to Sentinel-1 VV band images.
        Masks pixels with values less than -30.0 dB.

        Parameters:
        - image: ee.Image to be masked.

        Returns:
        - Masked ee.Image.
        """
        edge = image.lt(-30.0)
        masked_image = image.mask().And(edge.Not())
        return image.updateMask(masked_image)

    def apply_cloud_mask_hls(self, image):
        """
        Apply cloud mask for HLS using the Fmask band.
        """
        fmask = image.select('Fmask')
        return image.updateMask(fmask.eq(0))

    def apply_cloud_mask_s2(self, image):
        """
        Apply cloud mask for Sentinel-2 using the SCL band.
        """
        scl = image.select('SCL')
        # Remap classes: 4=Vegetation, 5=Bare Soil, 6=Water, 7=Unclassified, 11=Snow
        mask = scl.remap([4, 5, 6, 7, 11], [1, 1, 1, 1, 1], 0)
        return image.updateMask(mask.eq(1))

    def apply_cloud_mask_sentinel1(self, image):
        """
        Sentinel-1 is radar data and not affected by clouds.
        Apply edge masking using the provided mask_edge function.
        """
        # Apply edge masking
        image = self.mask_edge(image)
        return image

    def apply_cloud_mask(self, image):
        """
        Apply cloud masking depending on the data source.
        """
        if self.data_source == 'Sentinel-1':
            return self.apply_cloud_mask_sentinel1(image)
        elif self.data_source == 'HLS':
            return self.apply_cloud_mask_hls(image)
        elif self.data_source == 'Sentinel-2':
            return self.apply_cloud_mask_s2(image)
        else:
            logging.warning(f"Unknown data source '{self.data_source}'. No cloud masking applied.")
            return image

    def apply_qa_mask(self, image):
        """
        Placeholder for additional QA masking if needed.
        """
        return image

    def download_and_compute_indices(self):
        """
        Download satellite data and compute vegetation indices for each geometry for each available date.
        """
        # Get features as a list
        try:
            features = self.ee_feature_collection.getInfo()['features']
            logging.info(f"Processing {len(features)} features.")
        except Exception as e:
            logging.error(f"Failed to retrieve features from FeatureCollection: {e}")
            return

        total_features = len(features)
        for idx, feat in enumerate(features, start=1):
            feature_id = feat['properties'].get(self.field_id_col, 'unknown') if self.field_id_col else feat[
                'properties'].get('id', 'unknown')
            logging.info(f"Processing feature {idx}/{total_features}: '{feature_id}'")
            self.process_feature(feat)

    def process_feature(self, feature_info):
        """
        Process a single feature: For each available date in the given date range,
        compute vegetation indices and save them into a CSV.
        """
        feature = ee.Feature(feature_info)
        geom = feature.geometry()
        properties = feature_info['properties']

        # Retrieve start and end dates
        start_date = properties.get(self.start_date_col)
        end_date = properties.get(self.end_date_col)

        # Retrieve field identifier
        if self.field_id_col:
            field_identifier = properties.get(self.field_id_col, 'unknown')
        else:
            field_identifier = properties.get('id', 'unknown')

        feature_id = field_identifier if self.field_id_col else properties.get('id', 'unknown')

        if not start_date or not end_date:
            logging.warning(f"Skipping feature '{feature_id}': Missing start or end date.")
            return

        # Handle year-based data retrieval
        if start_date == end_date and len(start_date) == 4:
            try:
                year = int(start_date)
                if self.data_source == 'HLS' and year < 2018:
                    logging.warning(
                        f"Feature '{feature_id}': HLS data not available for year '{year}'. Skipping feature.")
                    return
                if self.data_source == 'Sentinel-1' and year < 2014:
                    logging.warning(
                        f"Feature '{feature_id}': Sentinel-1 data not available for year '{year}'. Skipping feature.")
                    return
                start_date = f"{year}-01-01"
                end_date = f"{year}-12-31"
                logging.info(
                    f"Feature '{feature_id}': Detected same start and end date as year '{year}'. Setting date range to entire year.")
            except ValueError:
                logging.warning(f"Feature '{feature_id}': Invalid year format '{start_date}'. Skipping feature.")
                return
        if year < 2019:
            logging.warning(f"Feature '{feature_id}': Data not available before 2019. Skipping feature.")
            return

        # Define image collection and bands based on data source
        if self.data_source == 'HLS':
            collection = ee.ImageCollection('NASA/HLS/HLSL30/v002').filterDate(start_date, end_date).filterBounds(geom)
            bands = ['B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'Fmask']
            scale = 30
        elif self.data_source == 'Sentinel-1':
            collection = ee.ImageCollection('COPERNICUS/S1_GRD') \
                .filterDate(start_date, end_date) \
                .filterBounds(geom) \
                .filter(ee.Filter.listContains('transmitterReceiverPolarisation', 'VV').Or(
                ee.Filter.listContains('transmitterReceiverPolarisation', 'HH')))
            bands = ['VV', 'VH']  # Initial assumption; will adjust based on available bands
            scale = 10
        elif self.data_source == 'Sentinel-2':
            collection = ee.ImageCollection('COPERNICUS/S2_SR_HARMONIZED').filterDate(start_date,
                                                                                      end_date).filterBounds(geom)
            bands = ['B2', 'B3', 'B4', 'B8', 'SCL']
            scale = 10
        else:
            logging.error(f"Unsupported data source '{self.data_source}'. Skipping feature '{feature_id}'.")
            return

        try:
            # Get collection size
            collection_size = collection.size().getInfo()
            logging.info(f"Feature '{feature_id}': Found {collection_size} images in the collection.")
            if collection_size == 0:
                logging.warning(f"No images found for feature '{feature_id}' within the specified date range.")
                return

            # Convert collection to a list for iteration
            image_list = collection.toList(collection_size)
            all_records = []

            for i in range(collection_size):
                img = ee.Image(image_list.get(i))

                # Detect if original geometry is a point and buffer if necessary
                geom_type = feature.geometry().type().getInfo()

                if geom_type == 'Point':
                    # Buffer by buffer_size meters
                    buffered_geom = geom.buffer(self.buffer_size)
                    logging.debug(
                        f"Feature '{feature_id}': Geometry is a Point. Buffered by {self.buffer_size} meters.")
                else:
                    buffered_geom = geom

                # Log buffered geometry bounds
                bounds = buffered_geom.bounds().getInfo()
                logging.debug(f"Feature '{feature_id}': Buffered Geometry Bounds: {bounds}")

                # Apply cloud and QA masks
                img = self.apply_cloud_mask(img)
                img = self.apply_qa_mask(img)

                # Adjust bands for Sentinel-1 based on available bands
                if self.data_source == 'Sentinel-1':
                    available_bands = img.bandNames().getInfo()
                    if 'VV' in available_bands and 'VH' in available_bands:
                        selected_bands = ['VV', 'VH']
                    elif 'HH' in available_bands and 'HV' in available_bands:
                        selected_bands = ['HH', 'HV']
                    else:
                        logging.error(
                            f"Feature '{feature_id}': Image does not contain required bands. Available bands: {available_bands}")
                        continue  # Skip this image

                    # Select relevant bands and clip to geometry
                    img = img.select(selected_bands).clip(buffered_geom)

                    # Compute indices based on available bands
                    indices_df = self.compute_vegetation_indices(img, buffered_geom, scale)
                    if indices_df.empty:
                        logging.warning(f"Feature '{feature_id}': No indices computed for image {i + 1}.")
                        continue
                else:
                    # Select relevant bands and clip to geometry
                    img = img.select(bands).clip(buffered_geom)

                    # Verify if selected bands exist in the image
                    available_bands = img.bandNames().getInfo()
                    missing_bands = [band for band in bands if band not in available_bands]
                    if missing_bands:
                        logging.error(
                            f"Feature '{feature_id}': Image does not contain bands {missing_bands}. Available bands: {available_bands}")
                        continue  # Skip this image

                    # Get image date
                    img_date = ee.Date(img.get('system:time_start')).format('YYYY-MM-DD').getInfo()
                    logging.debug(f"Feature '{feature_id}': Processing image dated {img_date}.")

                    # Compute indices for this single-date image
                    indices_df = self.compute_vegetation_indices(img, buffered_geom, scale)
                    if indices_df.empty:
                        logging.warning(f"Feature '{feature_id}': No indices computed for image {i + 1}.")
                        continue

                if not indices_df.empty:
                    # Add identification and date columns
                    indices_df['feature_id'] = feature_id
                    indices_df['field_id'] = properties.get(self.field_id_col,
                                                            'unknown') if self.field_id_col else properties.get('id',
                                                                                                                'unknown')
                    if self.data_source == 'Sentinel-1':
                        # Sentinel-1 images have dates; extract from metadata
                        img_date = ee.Date(img.get('system:time_start')).format('YYYY-MM-DD').getInfo()
                        indices_df['date'] = img_date
                    else:
                        indices_df['date'] = img_date
                    all_records.append(indices_df)

            if all_records:
                final_df = pd.concat(all_records, ignore_index=True)
                csv_name = f"{self.output_folder}/{feature_id}_{self.data_source}_{start_date}_{end_date}_indices.csv"
                final_df.to_csv(csv_name, index=False)
                logging.info(f"Processed and saved data for feature '{feature_id}'.")
            else:
                logging.warning(f"No valid data found for feature '{feature_id}' after processing.")
        except:
            logging.error(f"Failed to process feature '{feature_id}': {e}")

    def compute_vegetation_indices(self, image, geometry, scale):
        """
        Compute vegetation or radar indices for the provided image.
        """
        reducers = self.get_reducers()
        indices_list = []

        if self.data_source == 'HLS':
            # HLS band mapping: NIR=B5, RED=B4, GREEN=B3, BLUE=B2
            indices = {}
            indices['NDVI'] = image.normalizedDifference(['B5', 'B4']).rename('NDVI')
            indices['EVI'] = image.expression(
                '2.5 * ((NIR - RED) / (NIR + 6 * RED - 7.5 * BLUE + 1))', {
                    'NIR': image.select('B5'),
                    'RED': image.select('B4'),
                    'BLUE': image.select('B2')
                }).rename('EVI')
            indices['NDWI'] = image.normalizedDifference(['B3', 'B5']).rename('NDWI')
            indices['GNDVI'] = image.normalizedDifference(['B5', 'B3']).rename('GNDVI')
            indices['SAVI'] = image.expression(
                '((NIR - RED) / (NIR + RED + L)) * (1 + L)', {
                    'NIR': image.select('B5'),
                    'RED': image.select('B4'),
                    'L': 0.5
                }).rename('SAVI')
            indices['MSAVI'] = image.expression(
                '(2 * NIR + 1 - sqrt((2 * NIR + 1)**2 - 8 * (NIR - RED))) / 2', {
                    'NIR': image.select('B5'),
                    'RED': image.select('B4')
                }).rename('MSAVI')

        elif self.data_source == 'Sentinel-2':
            # Sentinel-2: NIR=B8, RED=B4, GREEN=B3, BLUE=B2
            indices = {}
            indices['NDVI'] = image.normalizedDifference(['B8', 'B4']).rename('NDVI')
            indices['EVI'] = image.expression(
                '2.5 * ((NIR - RED) / (NIR + 6 * RED - 7.5 * BLUE + 1))', {
                    'NIR': image.select('B8'),
                    'RED': image.select('B4'),
                    'BLUE': image.select('B2')
                }).rename('EVI')
            indices['NDWI'] = image.normalizedDifference(['B3', 'B8']).rename('NDWI')
            indices['GNDVI'] = image.normalizedDifference(['B8', 'B3']).rename('GNDVI')
            indices['SAVI'] = image.expression(
                '((NIR - RED) / (NIR + RED + L)) * (1 + L)', {
                    'NIR': image.select('B8'),
                    'RED': image.select('B4'),
                    'L': 0.5
                }).rename('SAVI')
            indices['MSAVI'] = image.expression(
                '(2 * NIR + 1 - sqrt((2 * NIR + 1)**2 - 8 * (NIR - RED))) / 2', {
                    'NIR': image.select('B8'),
                    'RED': image.select('B4')
                }).rename('MSAVI')

        elif self.data_source == 'Sentinel-1':
            # Sentinel-1: RVI computation handled based on available bands
            indices = {}
            available_bands = image.bandNames().getInfo()
            if 'VV' in available_bands and 'VH' in available_bands:
                indices['RVI'] = image.expression(
                    '4 * VH / (VV + VH)', {
                        'VV': image.select('VV'),
                        'VH': image.select('VH')
                    }).rename('RVI')
            elif 'HH' in available_bands and 'HV' in available_bands:
                indices['RVI'] = image.expression(
                    '4 * HV / (HH + HV)', {
                        'HH': image.select('HH'),
                        'HV': image.select('HV')
                    }).rename('RVI')
            else:
                # No valid bands for RVI
                logging.warning("Sentinel-1 image does not have the required bands for RVI computation.")
                return pd.DataFrame()

            # Optionally include individual bands
            if 'VV' in available_bands:
                indices['VV'] = image.select('VV')
            if 'VH' in available_bands:
                indices['VH'] = image.select('VH')
            if 'HH' in available_bands:
                indices['HH'] = image.select('HH')
            if 'HV' in available_bands:
                indices['HV'] = image.select('HV')

        else:
            # Unknown source, no indices
            logging.warning(f"Unknown data source '{self.data_source}'. No indices computed.")
            indices = {}

        # Compute statistics for each index
        for index_name, index_image in indices.items():
            stats_dict = index_image.reduceRegion(
                reducer=self.get_reducers(),
                geometry=geometry,
                scale=scale,
                bestEffort=True
            ).getInfo()
            if stats_dict:
                for stat_key, value in stats_dict.items():
                    indices_list.append({
                        'index': index_name,
                        'statistic': stat_key,
                        'value': value
                    })

        if indices_list:
            indices_df = pd.DataFrame(indices_list)
            return indices_df
        else:
            return pd.DataFrame()

    def get_reducers(self):
        """
        Create an ee.Reducer based on specified statistics.
        """
        reducers = []
        if 'mean' in self.stats:
            reducers.append(ee.Reducer.mean().setOutputs(['mean']))
        if 'median' in self.stats:
            reducers.append(ee.Reducer.median().setOutputs(['median']))
        if 'stdDev' in self.stats:
            reducers.append(ee.Reducer.stdDev().setOutputs(['stdDev']))
        if 'min' in self.stats:
            reducers.append(ee.Reducer.min().setOutputs(['min']))
        if 'max' in self.stats:
            reducers.append(ee.Reducer.max().setOutputs(['max']))
        if 'percentile' in self.stats:
            reducers.append(ee.Reducer.percentile(self.percentiles).setOutputs([f'p{p}' for p in self.percentiles]))

        if reducers:
            # Combine reducers into a single reducer
            combined_reducer = reducers[0]
            for r in reducers[1:]:
                combined_reducer = combined_reducer.combine(r, sharedInputs=True)
            return combined_reducer
        else:
            # Default to mean if no reducers are specified
            return ee.Reducer.mean().setOutputs(['mean'])

    def format_indices(self, index_name, stats_dict):
        """
        Format the indices into a list of dictionaries with 'index', 'statistic', 'value'.
        """
        formatted_list = []
        for stat_key, value in stats_dict.items():
            formatted_list.append({
                'index': index_name,
                'statistic': stat_key,
                'value': value
            })
        return formatted_list


# Example usage
if __name__ == "__main__":
    processor = SatelliteDataProcessor(
        shapefile_path=r'C:\Users\ritvik\Downloads\Yield_Afg_2015_2021 (2)\Yield_data_2015to2021_field_level.shp',  # Replace with your shapefile path
        start_date_col='Year',  # Replace with your start date column name
        end_date_col='Year',  # Replace with your end date column name
        data_source='Sentinel-2',  # Options: 'HLS', 'Sentinel-1', 'Sentinel-2'
        output_folder='downloads',
        stats=['mean', 'median'],
        percentiles=[25],
        field_id_col='Pnt_Id'  # Replace with your field identifier column name, if any
    )

    processor.download_and_compute_indices()



#!/usr/bin/env python

"""The setup script."""

import io
from os import path as op
from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

here = op.abspath(op.dirname(__file__))

# get the dependencies and installs
with io.open(op.join(here, "requirements.txt"), encoding="utf-8") as f:
    all_reqs = f.read().split("\n")

install_requires = [x.strip() for x in all_reqs if "git+" not in x]
dependency_links = [x.strip().replace("git+", "") for x in all_reqs if "git+" not in x]

requirements = []

setup_requirements = []

test_requirements = []

setup(
    author="Ritvik Sahajpal",
    author_email="ritvik@umd.edu",
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    description="Models to visualize and forecast crop conditions and yields",
    install_requires=install_requires,
    dependency_links=dependency_links,
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="geocif",
    name="geocif",
    packages=find_packages(include=["geocif", "geocif.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://ritviksahajpal.github.io/yield_forecasting/",
    version="0.1.65",
    zip_safe=False,
)
